/*----------------------------------------------------------------------
  File    : mlp.c
  Contents: training of very simple multilayer perceptrons
  Author  : Christian Borgelt
  History : 18.03.2000 file created from file xnn.c
            27.08.2000 renamed to mlp.c
            13.03.2001 bug in function mlp_train removed
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#include <math.h>
#include "mlpd.h"

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define ACT(x)    (1/(1 +exp(-(x))))
#define DERIV(x)  ((x)*(1-(x)))

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
/* --- training patterns --- */
static const PAT pats_xor[] = {    /* exclusive or */
  { 0, 0, 0 }, { 0, 1, 1 }, { 1, 0, 1 }, { 1, 1, 0 } };

static const PAT pats_biimp[] = {  /* biimplication */
  { 0, 0, 1 }, { 0, 1, 0 }, { 1, 0, 0 }, { 1, 1, 1 } };

static const PAT pats_fn1[] = {    /* function 1 */
  { -1.0, 0, 0.1  }, { -0.8, 0, 0.1  }, { -0.6, 0, 0.3  },
  { -0.4, 0, 0.5  }, { -0.2, 0, 0.7  }, {  0.0, 0, 0.9  },
  {  0.2, 0, 0.9  }, {  0.4, 0, 0.9  }, {  0.5, 0, 0.7  },
  {  0.6, 0, 0.5  }, {  0.7, 0, 0.3  }, {  0.8, 0, 0.1  },
  {  1.0, 0, 0.1  } };

static const PAT pats_fn2[] = {    /* function 2 */
  { -1.0, 0, 0.1  }, { -0.8, 0, 0.1  }, { -0.6, 0, 0.18 },
  { -0.4, 0, 0.26 }, { -0.2, 0, 0.34 }, {  0.0, 0, 0.42 },
  {  0.2, 0, 0.5  }, {  0.4, 0, 0.7  }, {  0.6, 0, 0.9  },
  {  0.8, 0, 0.9  }, {  1.0, 0, 0.9  } };

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
MLP mlp = {                     /* neural network */
  { { 0, 0, 0 }, { 0, 0, 0 }, { 0, 0, 0 } },   /* wgts */
  { { 0, 0, 0 }, { 0, 0, 0 }, { 0, 0, 0 } },   /* chgs */
  { 0, 0, 0 },                  /* out */
  MLP_XOR, MLP_BATCH,           /* type  and mode */
  0.2, 0.0,                     /* lrate and moment */
  0, 0,                         /* epoch and curr */
  { 2, sizeof(pats_xor)/sizeof(PAT), pats_xor }
};                              /* training pattern set */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/

void mlp_settype (int type)
{                               /* --- set problem type */
  mlp.type = type;              /* set the problem type */
  switch (type) {               /* evaluate the given type */
    case MLP_FUNC2:             /* -- function 2 */
      mlp.patset.incnt  = 1;    /* set patterns for function 2 */
      mlp.patset.patcnt = sizeof(pats_fn2)/sizeof(PAT);
      mlp.patset.pats   = pats_fn2;   break;
    case MLP_FUNC1:             /* -- function 1 */
      mlp.patset.incnt  = 1;    /* set patterns for function 1 */
      mlp.patset.patcnt = sizeof(pats_fn1)/sizeof(PAT);
      mlp.patset.pats   = pats_fn1;   break;
    case MLP_BIIMP:             /* -- biimplication */
      mlp.patset.incnt  = 2;    /* set patterns for biimplication */
      mlp.patset.patcnt = sizeof(pats_biimp)/sizeof(PAT);
      mlp.patset.pats   = pats_biimp; break;
    default:                    /* -- exclusive or */
      mlp.patset.incnt  = 2;    /* set patterns for exclusive or */
      mlp.patset.patcnt = sizeof(pats_xor)/sizeof(PAT);
      mlp.patset.pats   = pats_xor;
      mlp.type = MLP_XOR;             break;
  }                             /* ensure valid problem type code */
}  /* mlp_settype() */

/*--------------------------------------------------------------------*/

void mlp_init (double range)
{                               /* --- initialize neural network */
  int i;                        /* loop variable */

  if (range > -DBL_MAX) {       /* if to initialize the weights */
    for (i = 3; --i >= 0; ) {   /* choose random connection weights */
      mlp.wgts[i][0] = 2*range *(rand()/(double)RAND_MAX) -range;
      mlp.wgts[i][1] = 2*range *(rand()/(double)RAND_MAX) -range;
      if (mlp.type >= MLP_FUNC1)/* if real-valued function */
        mlp.wgts[i][2] = 2*range *(rand()/(double)RAND_MAX) -range;
      else {                    /* choose random weight, otherwise */
        mlp.wgts[i][2] = -(rand()/(double)RAND_MAX) *mlp.wgts[i][0]
                         -(rand()/(double)RAND_MAX) *mlp.wgts[i][1];
      }                         /* ensure that the line represented */
    }                           /* by each neuron is visible */
  }
  if (mlp.patset.incnt < 2)     /* if a real-valued function, clear */
    mlp.wgts[0][1] = mlp.wgts[1][1] = 0;  /* second input weights */
  mlp.chgs[0][0] = mlp.chgs[0][1] = mlp.chgs[0][2] = 0;
  mlp.chgs[1][0] = mlp.chgs[1][1] = mlp.chgs[1][2] = 0;
  mlp.chgs[2][0] = mlp.chgs[2][1] = mlp.chgs[2][2] = 0;
  mlp.epoch = 0;                /* clear weight changes and */
  mlp.curr  = 0;                /* init. epoch and current pattern */
}  /* mlp_init() */

/*--------------------------------------------------------------------*/

void mlp_train (void)
{                               /* --- do one training step */
  int     last;                 /* index of last training pattern */
  double  x,  y;                /* transformed coordinates */
  double  z,  ez;               /* network output and its error */
  double  ex, ey;               /* errors in hidden layer */
  double  d, t;                 /* temporary buffers */
  const PAT *p;                 /* to traverse the training patterns */

  mlp.chgs[0][0] *= mlp.moment;
  mlp.chgs[0][1] *= mlp.moment;
  mlp.chgs[0][2] *= mlp.moment; /* first  input neuron */
  mlp.chgs[1][0] *= mlp.moment;
  mlp.chgs[1][1] *= mlp.moment;
  mlp.chgs[1][2] *= mlp.moment; /* second input neuron */
  mlp.chgs[2][0] *= mlp.moment;
  mlp.chgs[2][1] *= mlp.moment;
  mlp.chgs[2][2] *= mlp.moment; /* output neuron */
  if (mlp.curr >= mlp.patset.patcnt) {
    mlp.curr = 0; mlp.epoch++; }
  last = (mlp.mode == MLP_BATCH)   /* get index of last pattern */
       ? mlp.patset.patcnt-1 : mlp.curr;
  p = mlp.patset.pats +mlp.curr;
  for (; mlp.curr <= last; p++, mlp.curr++) {
    t = mlp.wgts[0][0] *p->in1 +mlp.wgts[0][1] *p->in2 +mlp.wgts[0][2];
    x = ACT(t);
    t = mlp.wgts[1][0] *p->in1 +mlp.wgts[1][1] *p->in2 +mlp.wgts[1][2];
    y = ACT(t);
    t = mlp.wgts[2][0] *x      +mlp.wgts[2][1] *y      +mlp.wgts[2][2];
    z = ACT(t);                 /* compute the network output */
    ez = p->out -z;             /* and its error */
    mlp.chgs[2][2] += d = mlp.lrate *(t = ez *DERIV(z));
    mlp.chgs[2][1] += d *y;
    mlp.chgs[2][0] += d *x;
    ey = t *mlp.wgts[2][1];     /* backpropagate error for y */
    mlp.chgs[1][2] += d = mlp.lrate *(ey *DERIV(y));
    mlp.chgs[1][1] += d *p->in2;
    mlp.chgs[1][0] += d *p->in1;
    ex = t *mlp.wgts[2][0];     /* backpropagate error for x */
    mlp.chgs[0][2] += d = mlp.lrate *(ex *DERIV(x));
    mlp.chgs[0][1] += d *p->in2;
    mlp.chgs[0][0] += d *p->in1;
  }
  mlp.wgts[2][2] += mlp.chgs[2][2];
  mlp.wgts[2][1] += mlp.chgs[2][1];
  mlp.wgts[2][0] += mlp.chgs[2][0]; /* update output neuron */
  mlp.wgts[1][2] += mlp.chgs[1][2];
  mlp.wgts[1][1] += mlp.chgs[1][1];
  mlp.wgts[1][0] += mlp.chgs[1][0]; /* update second input neuron */
  mlp.wgts[0][2] += mlp.chgs[0][2];
  mlp.wgts[0][1] += mlp.chgs[0][1];
  mlp.wgts[0][0] += mlp.chgs[0][0]; /* update first  input neuron */
}  /* mlp_train() */

/*--------------------------------------------------------------------*/

double mlp_calc (double in1, double in2)
{                               /* --- calculate network output */
  double x, y, z;               /* outputs of the three neurons */

  x = mlp.wgts[0][0] *in1 +mlp.wgts[0][1] *in2 +mlp.wgts[0][2];
  mlp.out[0] = x = ACT(x);
  y = mlp.wgts[1][0] *in1 +mlp.wgts[1][1] *in2 +mlp.wgts[1][2];
  mlp.out[1] = y = ACT(y);
  z = mlp.wgts[2][0] *x   +mlp.wgts[2][1] *y   +mlp.wgts[2][2];
  mlp.out[2] = z = ACT(z);      /* calculate and return */
  return z;                     /* the network output */
}  /* mlp_calc() */
