/*----------------------------------------------------------------------
  File    : mlp.h
  Contents: training of very simple multilayer perceptrons
  Author  : Christian Borgelt
  History : 18.03.2000 file created from file xnn.c
            27.08.2000 renamed to mlp.h
----------------------------------------------------------------------*/
#ifndef __MLPD__
#define __MLPD__

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
/* --- neuron ids --- */
#define MLP_HD1        0        /* id of first  hidden neuron */
#define MLP_HD2        1        /* id of second hidden neuron */
#define MLP_OUT        2        /* id of output neuron */

/* --- connection ids --- */
#define MLP_IN1        0        /* first  input */
#define MLP_IN2        1        /* second input */
#define MLP_BIAS       2        /* (negated) bias */

/* --- problem types --- */
#define MLP_XOR        0        /* exclusive or */
#define MLP_BIIMP      1        /* biimplication */
#define MLP_FUNC1      2        /* function 1 */
#define MLP_FUNC2      3        /* function 2 */

/* --- learning mode --- */
#define MLP_ONLINE     0        /* update after each pattern */
#define MLP_BATCH      1        /* update after all patterns */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- training pattern --- */
  double in1, in2;              /* input values (in2 may be unused) */
  double out;                   /* output value */
} PAT;                          /* (training pattern) */

typedef struct {                /* --- training pattern set --- */
  int    incnt;                 /* number of inputs */
  int    patcnt;                /* number of training patterns */
  const PAT *pats;              /* vector of training patterns */
} PATSET;                       /* (training pattern set) */

typedef struct {                /* --- multilayer perceptron --- */
  double wgts[3][3];            /* connection weights */
  double chgs[3][3];            /* weight changes */
  double out [3];               /* neuron outputs */
  int    type;                  /* problem type,  e.g. MLP_XOR */
  int    mode;                  /* learning mode, e.g. MLP_BATCH */
  double lrate;                 /* learning rate */
  double moment;                /* momentum */
  int    epoch;                 /* current training epoch */
  int    curr;                  /* current input pattern */
  PATSET patset;                /* training pattern set */
} MLP;                          /* (multilayer perceptron) */

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
extern MLP mlp;                 /* multilayer perceptron */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/
extern void   mlp_settype   (int type);
extern int    mlp_gettype   (void);
extern void   mlp_setmode   (int mode);
extern int    mlp_getmode   (void);
extern void   mlp_setlrate  (double lrate);
extern double mlp_getlrate  (void);
extern void   mlp_setmoment (double moment);
extern double mlp_getmoment (void);

extern void   mlp_init      (double range);
extern void   mlp_setwgt    (int neuron, int input, double wgt);
extern double mlp_getwgt    (int neuron, int input);

extern int    mlp_epoch     (void);
extern void   mlp_train     (void);
extern double mlp_calc      (double in1, double in2);
extern double mlp_output    (int neuron);

extern int    mlp_patcnt    (void);
extern const PAT* mlp_pattern (int index);

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define mlp_gettype()       (mlp.type)
#define mlp_setmode(m)      (mlp.mode   = (m))
#define mlp_getmode()       (mlp.mode)
#define mlp_setlrate(r)     (mlp.lrate  = (r))
#define mlp_getlrate()      (mlp.lrate)
#define mlp_setmoment(m)    (mlp.moment = (m))
#define mlp_getmoment()     (mlp.moment)

#define mlp_setwgt(u,i,w)   (mlp.wgts[u][i] = (w))
#define mlp_getwgt(u,i)     (mlp.wgts[u][i])

#define mlp_epoch()         (mlp.epoch)
#define mlp_output(u)       (mlp.out[u])

#define mlp_patcnt()        (mlp.patset.patcnt)
#define mlp_pattern(i)      (mlp.patset.pats +i)

#endif
