/*----------------------------------------------------------------------
  File    : wmlp.c
  Contents: Windows program to visualize multilayer perceptron training
  Author  : Christian Borgelt
  History : 29.10.1999 file created from file xnn.c
            30.10.1999 first version completed
            09.11.1999 minor improvements
            18.03.2000 adapted to module mlp, function approx. added
            27.08.2000 renamed to wmlp.c
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <float.h>
#include <math.h>
#include <time.h>
#include <windows.h>
#include "wmlp.rh"
#include "mlpd.h"

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME  "wmlp"         /* main window title */

/* --- abbreviations --- */
#define HINST             HINSTANCE
#define MKIRSC            MAKEINTRESOURCE
#define SDIMSG            SendDlgItemMessage
#define WINSTYLE          WS_CAPTION|WS_SYSMENU|WS_BORDER|WS_MINIMIZEBOX
#define PATH_MAX          _MAX_PATH

/* --- edit box functions --- */
#define edt_set(h,i,s)    SetDlgItemText(h, i, s)
#define edt_get(h,i,s,n)  GetDlgItemText(h, i, s, n)
#define int_set(h,i,n)    SetDlgItemInt(h, i, n, 0)
#define int_get(h,i)      GetDlgItemInt(h, i, NULL, 0)

/* --- combo box functions --- */
#define cmb_add(h,i,s)    SDIMSG(h, i, CB_ADDSTRING,    0, (LPARAM)(s))
#define cmb_rem(h,i,k)    SDIMSG(h, i, CB_DELETESTRING, k, 0)
#define cmb_set(h,i,k)    SDIMSG(h, i, CB_SETCURSEL,    k, 0)
#define cmb_get(h,i)      SDIMSG(h, i, CB_GETCURSEL,    0, 0)

/* --- extensions --- */
#define INITXEXT     572        /* initial x-extension of window */
#define INITYEXT     400        /* initial y-extension of window */
#define MINVAL       (-0.2)     /* range of values */
#define MAXVAL         1.2      /* for diagrams */

/* --- error codes --- */
#define OK             0        /* no error */
#define E_NONE         0        /* no error */
#define E_NOMEM      (-1)       /* not enough memory */
#define E_FOPEN      (-2)       /* cannot open file */
#define E_FREAD      (-3)       /* read error on file */
#define E_FWRITE     (-4)       /* write error on file */
#define E_INIT       (-5)       /* initialization failed */
#define E_UNKNOWN    (-6)       /* unkown error */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
const char *errmsgs[] = {       /* error messages */
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file\n%s",
  /* E_FREAD    -3 */  "read error on file\n%s",
  /* E_FWRITE   -4 */  "write error on file\n%s",
  /* E_INIT     -5 */  "initialization failed",
  /* E_UNKNOWN  -6 */  "unknown error"
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static HINST   hinst;           /* instance handle */
static HWND    hmain;           /* handle of main window */
static HDC     hdc_win;         /* device context of window */
static HDC     hdc_buf;         /* device context of drawing buffer */
static HBITMAP buffer;          /* drawing buffer */
static char    text[_MAX_PATH +1024];

/* --- pens --- */
static HPEN    pen_black;       /* black  : coordinate axes */
static HPEN    pen_red;         /* yellow : pattern (1,1) */
static HPEN    pen_green;       /* green  : pattern (0,1) */
static HPEN    pen_blue;        /* blue   : pattern (1,0) */
static HPEN    pen_yellow;      /* yellow : pattern (0,0) */
static HPEN    pen_grey;        /* grey   : auxiliary lines */
static HPEN    pen_magenta;     /* magenta: upper hidden neuron */
static HPEN    pen_cyan;        /* cyan   : lower hidden neuron */
static HPEN    pen_orange;      /* orange : output neuron */

/* --- brushes --- */
static HBRUSH  br_white;        /* white  : background */
static HBRUSH  br_black;        /* black  : arrow heads */
static HBRUSH  br_red;          /* yellow : pattern (1,1) */
static HBRUSH  br_green;        /* green  : pattern (0,1) */
static HBRUSH  br_blue;         /* blue   : pattern (1,0) */
static HBRUSH  br_yellow;       /* yellow : pattern (0,0) */
static HBRUSH  br_grey;         /* grey   : input neurons */
static HBRUSH  br_magenta;      /* magenta: upper hidden neuron */
static HBRUSH  br_cyan;         /* cyan   : lower hidden neuron */
static HBRUSH  br_orange;       /* orange : output neuron */

/* --- neural network variables --- */
static int     xoff, yoff;      /* offsets for drawing */
static double  range   = 1;     /* initial weight range */
static double  delay   = 0.01;  /* delay between training steps */
static int     diag    = 0;     /* whether diagram has been drawn */
static int     running = 0;     /* whether training is running */

/*----------------------------------------------------------------------
  Error Functions
----------------------------------------------------------------------*/

static void error (int code, ...)
{                               /* --- print error message */
  va_list     args;             /* list of variable arguments */
  const char  *msg;             /* error message */

  if ((code > 0) || (code < E_UNKNOWN))
    code = E_UNKNOWN;           /* check error code */
  va_start(args, code);         /* get variable arguments */
  msg = errmsgs[-code];         /* and error message text */
  if (!msg) msg = errmsgs[-E_UNKNOWN];
  vsprintf(text, msg, args);    /* format error message */
  MessageBox(hmain, text, PRGNAME, MB_OK|MB_ICONSTOP);
  va_end(args);                 /* show alert dialog box */
}  /* error() */

/*----------------------------------------------------------------------
  Drawing Functions
----------------------------------------------------------------------*/

static void arrow (int x1, int y1, int x2, int y2)
{                               /* --- draw a horz. or vert. arrow */
  MoveToEx(hdc_buf, x1, y1, NULL);
  if (x1 == x2) {               /* if vertical arrow */
    LineTo(hdc_buf, x2, y2+5);
    Pie   (hdc_buf, x2-6, y2+6, x2+7, y2-6, x2-2, y2+6, x2+3, y2+6); }
  else {                        /* if horizontal arrow */
    LineTo(hdc_buf, x2-6, y2);  
    Pie   (hdc_buf, x2-7, y2+7, x2+6, y2-6, x2-8, y2-2, x2-8, y2+3);
  }
}  /* arrow() */

/*--------------------------------------------------------------------*/

static void conwgt (int x, int y, double wgt)
{                               /* --- draw a connection weight */
  RECT rc;                      /* rectangle for formatting */
  int  len, wd;                 /* string length and text width */
  char s[32];                   /* buffer for weight as a text */

  len = sprintf(s," %.4g ",wgt);/* convert the weight to a string*/
  rc.left = rc.right  = x;
  rc.top  = rc.bottom = y;      /* compute the text extensions */
  wd = DrawText(hdc_buf, s, len, &rc, DT_CALCRECT);
  rc.top  -= wd/2; rc.bottom -= wd/2;
  wd = rc.right -rc.left;       /* center the text to the point */
  rc.left -= wd/2; rc.right  -= wd/2;
  DrawText(hdc_buf, s, len, &rc, DT_CENTER|DT_VCENTER);
}  /* conwgt() */               /* draw the connection weight */

/*--------------------------------------------------------------------*/

static void point (double x, double y, int value,
                   int diag, int size, HPEN pen, HBRUSH brush)
{                               /* --- draw a point */
  int wx = xoff +(diag > 0) ? 328 : 44;
  int wy = yoff -44;            /* window coordinates */

  wx += (int)(x *200) -size/2;  /* compute window coordinates */
  wy -= (int)(y *200) +size/2;  /* of the point to draw */
  SelectObject(hdc_buf, pen);   /* get pen and brush */
  SelectObject(hdc_buf, brush); /* of the right color */
  Ellipse(hdc_buf, wx, wy, wx +size, wy +size);
  if (value == 0) {             /* if the output value is zero */
    SelectObject(hdc_buf, br_white);
    Ellipse(hdc_buf, wx+1, wy+1, wx +size-1, wy +size-1);
  }                             /* make it a circle */
}  /* point() */

/*--------------------------------------------------------------------*/

static void line (double a, double b, double c, int diag,
                  HPEN pen, HBRUSH brush)
{                               /* --- draw a line */
  int    wx = xoff +(diag > 0) ? 328 : 44;
  int    wy = yoff -44;         /* window coordinates */
  double x1, y1, x2, y2;        /* coordinates for lines */
  double t;                     /* temporary buffer */

  /* --- clipping --- */
  if (fabs(a) < 1e-12) a = (a < 0) ? -1e-12 : 1e-12;
  if (fabs(b) < 1e-12) b = (b < 0) ? -1e-12 : 1e-12;
  x1 = (-MINVAL *b -c) /a;
  if      (x1 < MINVAL) { x1 = MINVAL; y1 = (-MINVAL *a -c) /b; }
  else if (x1 > MAXVAL) { x1 = MAXVAL; y1 = (-MAXVAL *a -c) /b; }
  else                  {              y1 = MINVAL; }
  x2 = (-MAXVAL *b -c) /a;
  if      (x2 < MINVAL) { x2 = MINVAL; y2 = (-MINVAL *a -c) /b; }
  else if (x2 > MAXVAL) { x2 = MAXVAL; y2 = (-MAXVAL *a -c) /b; }
  else                  {              y2 = MAXVAL; }
  if ((y1 < MINVAL) || (y1 > MAXVAL)
  ||  (y2 < MINVAL) || (y2 > MAXVAL))
    return;

  /* --- drawing --- */
  SelectObject(hdc_buf, pen);   /* get pen and brush */
  SelectObject(hdc_buf, brush); /* of the right color */
  MoveToEx(hdc_buf, wx +(int)(x1 *200), wy -(int)(y1 *200), NULL);
  LineTo  (hdc_buf, wx +(int)(x2 *200), wy -(int)(y2 *200));
  t  = sqrt(a*a + b*b);
  x1 = (x1 +x2) /2; x2 = x1 +0.05 *(a /= t);
  y1 = (y1 +y2) /2; y2 = y1 +0.05 *(b /= t);
  MoveToEx(hdc_buf, wx +(int)(x1 *200), wy -(int)(y1 *200), NULL);
  LineTo  (hdc_buf, wx +(int)(x2 *200), wy -(int)(y2 *200));
  wx += (int)(x2 *200);
  wy -= (int)(y2 *200);
  x1 = -a -0.35 *b; x2 = -a +0.35 *b;
  y1 = -b +0.35 *a; y2 = -b -0.35 *a;
  Pie(hdc_buf, wx-6, wy+7, wx+7, wy-6,
               wx +(int)(x1 *200), wy -(int)(y1 *200),
               wx +(int)(x2 *200), wy -(int)(y2 *200));
}  /* line() */

/*--------------------------------------------------------------------*/

static void patterns (void)
{                               /* --- draw training patterns */
  int  cnt;                     /* number of patterns */
  int  x0 = xoff +INITXEXT/2;   /* x-coordinate of origin */
  int  y0 = yoff -28;           /* y-coordinate of origin */
  RECT rc;                      /* rectangle for drawing */
  const PAT *p;                 /* to traverse the patterns */
  
  cnt = mlp_patcnt();           /* get the number of patterns */
  while (--cnt >= 0) {          /* and traverse the patterns */
    p = mlp_pattern(cnt);       /* get the next training pattern */
    rc.left = x0 +(int)(230 *p->in1)-1; rc.right  = rc.left +3;
    rc.top  = y0 -(int)(230 *p->out)-1; rc.bottom = rc.top  +3;
    FillRect(hdc_buf, &rc, br_black);
  }                             /* draw circles at the data points */
}  /* patterns() */

/*--------------------------------------------------------------------*/

static void netout (void)
{                               /* --- draw network output */
  static int hd1[101];          /* hidden neuron 1 output */
  static int hd2[101];          /* hidden neuron 2 output */
  static int out[101];          /* output neuron output */
  int    x0 = xoff +36;         /* x-coordinate of reference point */
  int    y0 = yoff -28;         /* y-coordinate of reference point */
  int    x;                     /* loop variable, pixel coordinate */

  for (x = 101; --x >= 0; ) {   /* traverse the input range */
    out[x] = y0 -(int)(230 *mlp_calc((x-50)*(5/230.0), 0));
    hd1[x] = y0 -(int)(230 *mlp_output(0));
    hd2[x] = y0 -(int)(230 *mlp_output(1));
  }                             /* compute outputs of neurons */
  SelectObject(hdc_buf, pen_blue);
  MoveToEx(hdc_buf, x0+500, hd1[100], NULL);
  for (x = 100; --x >= 0; ) LineTo(hdc_buf, x0+5*x, hd1[x]);
  SelectObject(hdc_buf, pen_green);
  MoveToEx(hdc_buf, x0+500, hd2[100], NULL);
  for (x = 100; --x >= 0; ) LineTo(hdc_buf, x0+5*x, hd2[x]);
  SelectObject(hdc_buf, pen_red);
  MoveToEx(hdc_buf, x0+500, out[100], NULL);
  for (x = 100; --x >= 0; ) LineTo(hdc_buf, x0+5*x, out[x]);
}  /* netout() */

/*--------------------------------------------------------------------*/

static void redraw (HDC hdc)
{                               /* --- Actions > Redraw */
  RECT rc;                      /* rectangle for drawing */
  int  type;                    /* problem type */
  int  bool;                    /* Boolean function flag */
  const PAT *p;                 /* to traverse the training patterns */

  type = mlp_gettype();         /* get the problem type and */
  bool = (type < MLP_FUNC1);    /* compute Boolean function flag */
  if (diag) {                   /* if diagram has been drawn once */
    rc.left = 0;        rc.right  = INITXEXT-1;
    rc.top  = yoff-294; rc.bottom = yoff;
    FillRect(hdc_buf, &rc, br_white);
    rc.left = xoff+124; rc.right  = xoff+264;
    rc.top  = yoff-389; rc.bottom = yoff-298;
    FillRect(hdc_buf, &rc, br_white);
    rc.left = xoff+299; rc.right  = xoff+439;
    rc.top  = yoff-389; rc.bottom = yoff-298;
    FillRect(hdc_buf, &rc, br_white);
    rc.left = xoff+400; rc.right  = xoff+530;
    rc.top  = yoff-332; rc.bottom = yoff-298;
    FillRect(hdc_buf, &rc, br_white);
    rc.left = xoff +((bool) ? 90 : 264); rc.right = xoff+299;
    rc.top  = yoff-366; rc.bottom = yoff-321;
    FillRect(hdc_buf, &rc, br_white); }
  else {                        /* if diagram has not been drawn yet */
    rc.left = 0; rc.right  = INITXEXT-1;
    rc.top  = 0; rc.bottom = INITYEXT-1;
    FillRect(hdc_buf, &rc, br_white);

    /* --- neurons --- */
    SelectObject(hdc_buf, pen_black);
    SelectObject(hdc_buf, br_grey);
    if (!bool)                  /* one input for real-valued function */
      Ellipse(hdc_buf, xoff+100, yoff-333, xoff+121, yoff-354);
    else {                      /* two inputs for Boolean function */
      Ellipse(hdc_buf, xoff+100, yoff-368, xoff+121, yoff-389);
      Ellipse(hdc_buf, xoff+100, yoff-298, xoff+121, yoff-319);
    }
    SelectObject(hdc_buf, (bool) ? br_cyan    : br_blue);
    Ellipse(hdc_buf, xoff+275, yoff-298, xoff+296, yoff-319);
    SelectObject(hdc_buf, (bool) ? br_magenta : br_green);
    Ellipse(hdc_buf, xoff+275, yoff-368, xoff+296, yoff-389);
    SelectObject(hdc_buf, (bool) ? br_orange  : br_red);
    Ellipse(hdc_buf, xoff+450, yoff-333, xoff+471, yoff-354);
    SelectObject(hdc_buf, br_black);

    /* --- connections --- */
    if (bool) {                 /* two inputs for Boolean functions */
      arrow(xoff +40, yoff-379, xoff +97, yoff-379);  /* upper */
      arrow(xoff+110, yoff-372, xoff+110, yoff-385);  /* input */
      arrow(xoff +40, yoff-309, xoff +97, yoff-309);  /* lower */
      arrow(xoff+103, yoff-309, xoff+118, yoff-309);  /* input */
      arrow(xoff+124, yoff-379, xoff+272, yoff-379);  /* upper */
      arrow(xoff+285, yoff-372, xoff+285, yoff-385);  /* hidden */
      Pie(hdc_buf, xoff+266, yoff-309, xoff+282, yoff-323,
                   xoff+266, yoff-323, xoff+266, yoff-317);
      arrow(xoff+124, yoff-309, xoff+272, yoff-309);  /* lower */
      arrow(xoff+278, yoff-309, xoff+293, yoff-309);  /* hidden */
      Pie(hdc_buf, xoff+266, yoff-377, xoff+282, yoff-363,
                   xoff+266, yoff-371, xoff+266, yoff-366); }
    else {                      /* one input for real-valued function */
      arrow(xoff +40, yoff-344, xoff +97, yoff-344);  /* input */
      Pie(hdc_buf, xoff+264, yoff-385, xoff+280, yoff-371,
                   xoff+264, yoff-379, xoff+264, yoff-374);
      Pie(hdc_buf, xoff+264, yoff-301, xoff+280, yoff-315,
                   xoff+264, yoff-315, xoff+264, yoff-309);
    }
    Pie(hdc_buf, xoff+439, yoff-346, xoff+455, yoff-332,
                 xoff+439, yoff-340, xoff+439, yoff-335);
    Pie(hdc_buf, xoff+439, yoff-339, xoff+455, yoff-356,
                 xoff+439, yoff-352, xoff+439, yoff-348);
    arrow(xoff+474, yoff-344, xoff+531, yoff-344);
    diag = 1;                   /* set diagram initialization flag */
  }

  /* -- connections -- */
  SelectObject(hdc_buf, pen_black);
  if (bool) {                   /* if Boolean function */
    MoveToEx(hdc_buf, xoff+124, yoff-379, NULL);  /* upper */
    LineTo  (hdc_buf, xoff+272, yoff-379);        /* hidden */
    MoveToEx(hdc_buf, xoff+122, yoff-317, NULL);
    LineTo  (hdc_buf, xoff+274, yoff-371);
    MoveToEx(hdc_buf, xoff+124, yoff-309, NULL);  /* lower */
    LineTo  (hdc_buf, xoff+272, yoff-309);        /* hidden */
    MoveToEx(hdc_buf, xoff+122, yoff-371, NULL);
    LineTo  (hdc_buf, xoff+274, yoff-317); }
  else {                        /* if real-valued function */
    MoveToEx(hdc_buf, xoff+124, yoff-340, NULL);
    LineTo  (hdc_buf, xoff+272, yoff-309);
    MoveToEx(hdc_buf, xoff+124, yoff-348, NULL);
    LineTo  (hdc_buf, xoff+272, yoff-379);
  }                             /* (connections to hidden neurons) */
  MoveToEx(hdc_buf, xoff+299, yoff-379, NULL);
  LineTo  (hdc_buf, xoff+447, yoff-348);
  MoveToEx(hdc_buf, xoff+299, yoff-309, NULL);
  LineTo  (hdc_buf, xoff+447, yoff-340);

  /* --- connection weights --- */
  if (bool) {                   /* if Boolean function */
    conwgt(xoff+198, yoff-309, mlp_getwgt(MLP_HD1, MLP_IN1));
    conwgt(xoff+158, yoff-357, mlp_getwgt(MLP_HD1, MLP_IN2));
    conwgt(xoff+158, yoff-330, mlp_getwgt(MLP_HD2, MLP_IN1));
    conwgt(xoff+198, yoff-379, mlp_getwgt(MLP_HD2, MLP_IN2)); }
  else {                        /* if real-valued function */
    conwgt(xoff+198, yoff-324, mlp_getwgt(MLP_HD1, MLP_IN1));
    conwgt(xoff+198, yoff-364, mlp_getwgt(MLP_HD2, MLP_IN1));
  }                             /* (connections to hidden neurons) */
  conwgt(xoff+286, yoff-332, -mlp_getwgt(MLP_HD1, MLP_BIAS));
  conwgt(xoff+286, yoff-355, -mlp_getwgt(MLP_HD2, MLP_BIAS));
  conwgt(xoff+373, yoff-324,  mlp_getwgt(MLP_OUT, MLP_IN1));
  conwgt(xoff+373, yoff-364,  mlp_getwgt(MLP_OUT, MLP_IN2));
  conwgt(xoff+461, yoff-321, -mlp_getwgt(MLP_OUT, MLP_BIAS));

  /* --- separating line --- */
  SelectObject(hdc_buf, pen_grey);
  MoveToEx(hdc_buf, xoff,              yoff-288, NULL);
  LineTo  (hdc_buf, xoff +INITXEXT -1, yoff-288);

  if (bool) {                   /* if Boolean function */
    /* --- left diagram --- */
    SelectObject(hdc_buf, pen_grey);
    MoveToEx(hdc_buf, xoff+244, yoff -44, NULL);
    LineTo  (hdc_buf, xoff+244, yoff-244);
    LineTo  (hdc_buf, xoff +44, yoff-244);
    SelectObject(hdc_buf, pen_black);
    SelectObject(hdc_buf, br_black);
    arrow(xoff +44, yoff -44, xoff +44, yoff-264);
    arrow(xoff +44, yoff -44, xoff+265, yoff -44);
    p = mlp_pattern(0);
    point(p->in1, p->in2, (int)p->out, 0, 9, pen_yellow, br_yellow);
    p = mlp_pattern(1);
    point(p->in1, p->in2, (int)p->out, 0, 9, pen_green,  br_green);
    p = mlp_pattern(2);
    point(p->in1, p->in2, (int)p->out, 0, 9, pen_blue,   br_blue);
    p = mlp_pattern(3);
    point(p->in1, p->in2, (int)p->out, 0, 9, pen_red,    br_red);
    line(mlp_getwgt(MLP_HD1, MLP_IN1), mlp_getwgt(MLP_HD1, MLP_IN2),
         mlp_getwgt(MLP_HD1, MLP_BIAS), 0, pen_cyan,    br_cyan);
    line(mlp_getwgt(MLP_HD2, MLP_IN1), mlp_getwgt(MLP_HD2, MLP_IN2),
         mlp_getwgt(MLP_HD2, MLP_BIAS), 0, pen_magenta, br_magenta);

    /* --- right diagram --- */
    SelectObject(hdc_buf, pen_grey);
    MoveToEx(hdc_buf, xoff+528, yoff -44, NULL);
    LineTo  (hdc_buf, xoff+528, yoff-244);
    LineTo  (hdc_buf, xoff+328, yoff-244);
    MoveToEx(hdc_buf, xoff+428, yoff -44, NULL);
    LineTo  (hdc_buf, xoff+428, yoff-244);
    MoveToEx(hdc_buf, xoff+328, yoff-144, NULL);
    LineTo  (hdc_buf, xoff+528, yoff-144);
    SelectObject(hdc_buf, pen_black);
    SelectObject(hdc_buf, br_black);
    arrow(xoff+328, yoff -44, xoff+328, yoff-264);
    arrow(xoff+328, yoff -44, xoff+549, yoff -44);
    p = mlp_pattern(0); mlp_calc(p->in1, p->in2);
    point(mlp_output(0), mlp_output(1), (int)p->out,
          1, 9, pen_yellow, br_yellow);
    p = mlp_pattern(1); mlp_calc(p->in1, p->in2);
    point(mlp_output(0), mlp_output(1), (int)p->out,
          1, 9, pen_green,  br_green);
    p = mlp_pattern(2); mlp_calc(p->in1, (int)p->in2);
    point(mlp_output(0), mlp_output(1), (int)p->out,
          1, 9, pen_blue,   br_blue);
    p = mlp_pattern(3); mlp_calc(p->in1, (int)p->in2);
    point(mlp_output(0), mlp_output(1), (int)p->out,
          1, 9, pen_red,    br_red);
    line(mlp_getwgt(MLP_OUT, MLP_IN1), mlp_getwgt(MLP_OUT, MLP_IN2),
         mlp_getwgt(MLP_OUT, MLP_BIAS), 1, pen_orange, br_orange);
    }
  else {                        /* if real-valued function */
    /* --- diagram --- */
    SelectObject(hdc_buf, pen_grey);
    MoveToEx(hdc_buf, xoff +20, yoff-258, NULL);
    LineTo  (hdc_buf, xoff+552, yoff-258);
    MoveToEx(hdc_buf, xoff +20, yoff-143, NULL);
    LineTo  (hdc_buf, xoff+552, yoff-143);
    SelectObject(hdc_buf, pen_black);
    SelectObject(hdc_buf, br_black);
    arrow(xoff+286, yoff -25, xoff+286, yoff-274);
    arrow(xoff+ 21, yoff -28, xoff+552, yoff -28);
    Pie(hdc_buf, xoff +12, yoff -21, xoff +25, yoff -34,
                 xoff +26, yoff -24, xoff +25, yoff -31);
    MoveToEx(hdc_buf, xoff+283, yoff-258, NULL);
    LineTo  (hdc_buf, xoff+290, yoff-258);
    MoveToEx(hdc_buf, xoff +56, yoff -25, NULL);
    LineTo  (hdc_buf, xoff +56, yoff -32);
    MoveToEx(hdc_buf, xoff+516, yoff -25, NULL);
    LineTo  (hdc_buf, xoff+516, yoff -32);
    patterns();                 /* draw the training data and */
    netout();                   /* the outputs of the neurons */
  }

  /* --- epoch --- */
  sprintf(text, "%s %10ld", PRGNAME, mlp_epoch());
  SetWindowText(hmain, text);

  /* --- copy buffer to screen --- */
  if (!hdc) hdc = hdc_win;      /* get device context of window */
  BitBlt(hdc, 0, 0, INITXEXT-1, INITYEXT-1, hdc_buf, 0, 0, SRCCOPY);
}  /* redraw() */

/*----------------------------------------------------------------------
  Neural Network Training Function
----------------------------------------------------------------------*/

static void init (int type)
{                               /* --- initialize neural network */
  mlp_settype(type);            /* get and set the problem type, */
  mlp_init(range);              /* initialize the network weights, */
  diag = 0;                     /* toggle a full redraw, and */
  SetWindowText(hmain, PRGNAME);/* reset the window title */
}  /* init() */

/*--------------------------------------------------------------------*/

static void CALLBACK train (HWND hwnd, UINT msg, UINT id, DWORD systime)
{                               /* --- do one training step */
  if (!running) return;         /* check running flag */
  mlp_train();                  /* train the neural network and */
  redraw(NULL);                 /* redraw the window contents */
  SetTimer(hwnd, 1, (UINT)(delay *1000), (TIMERPROC)train);
}  /* train() */                /* reinstall timer */

/*----------------------------------------------------------------------
  Dialog Functions
----------------------------------------------------------------------*/

static char* fselect (char *title, int load)
{                               /* --- select a file */
  static char fname[_MAX_PATH+1];    /* file name */
  static char path [_MAX_PATH+1];    /* and path */
  static char filter[] =        /* file selection filter */
    "network weights (*.mlp)\0*.mlp\0" /* neural network weights */
    "all files (*.*)\0*.*\0";          /* all files */
  static OPENFILENAME ofn = {   /* file selector structure */
    sizeof(OPENFILENAME),       /* lStructSize */
    NULL,   NULL,               /* hwndOwner, hInstance */
    filter, NULL,               /* lpstrFilter, lpstrCustomFilter */
    0, 1,                       /* nMaxCustFilter, nFilterIndex */
    fname, _MAX_PATH,           /* lpstrFile, nMaxFile */
    NULL, _MAX_FNAME+_MAX_EXT,  /* lpstrFileTitle, nMaxFileTitle */
    path,                       /* lpstrInitialDir */
    "Select File...",           /* lpstrTitle */
    OFN_PATHMUSTEXIST, 0, 0,    /* Flags, nFileOffset, nFileExtension */
    "mlp", 0L,                  /* lpstrDefExt, lCustData */
    NULL, NULL };               /* lpfnHook, lpTemplateName */

  ofn.hwndOwner  = hmain;       /* set window handle */
  ofn.lpstrTitle = title;       /* and title text */
  ofn.Flags = (load) ? OFN_FILEMUSTEXIST : OFN_PATHMUSTEXIST;
  if (GetOpenFileName(&ofn) == 0)
    return NULL;                /* open file selector box */
  return fname;                 /* return name of selected file */
}  /* fselect() */

/*--------------------------------------------------------------------*/

static int load (void)
{                               /* --- File > Load Weights */
  int    i;                     /* loop variable */
  char   *fname;                /* name of file to load */
  FILE   *file;                 /* input file */
  double w[3][3];               /* buffer for weights */

  LoadString(hinst, CS_LOAD, text, 80);
  fname = fselect(text, 1);     /* get the dialog box title */
  if (!fname) return -1;        /* and select a file */
  file  = fopen(fname, "r");    /* open input file */
  if (!file) { error(E_FOPEN, fname); return -1; }
  for (i = 0; i < 3; i++) {     /* read weights from file */
    if (fscanf(file, "%lf%lf%lf", w[i], w[i] +1, w[i] +2) != 3) {
      error(E_FREAD, fname); return -1; }
  }                             /* store weights in buffer */
  fclose(file);                 /* close input file */
  for (i = 3; --i >= 0; ) {     /* set the connection weights */
    mlp_setwgt(i, MLP_IN1,   w[i][0]);     /* from the buffer */
    mlp_setwgt(i, MLP_IN2,   w[i][1]);
    mlp_setwgt(i, MLP_BIAS, -w[i][2]);
  }
  mlp_init(-DBL_MAX);           /* initialize the neural network */
  return 0;                     /* return 'ok' */
}  /* load() */

/*--------------------------------------------------------------------*/

static void save (void)
{                               /* --- File > Save Weights... */
  int  i;                       /* loop variable */
  char *fname;                  /* name of file to save to */
  FILE *file;                   /* output file */

  LoadString(hinst, CS_SAVE, text, 80);
  fname = fselect(text, 0);     /* get the dialog box title */
  if (!fname) return;           /* and select a file */
  file = fopen(fname, "w");     /* open output file */
  if (!file) { error(E_FOPEN, fname); return; }
  for (i = 0; i < 3; i++)       /* write the connection weights */
    fprintf(file, "%g %g %g\n", /* to the chosen file */
            mlp_getwgt(i, MLP_IN1),
            mlp_getwgt(i, MLP_IN2),
           -mlp_getwgt(i, MLP_BIAS));
  if (fclose(file) != 0) error(E_FWRITE, fname);
}  /* save() */

/*--------------------------------------------------------------------*/

static void mn_enable (int enable)
{                               /* --- enable/disable menu items */
  HMENU        menu;            /* handle of window menu */
  MENUITEMINFO mii;             /* for menu customization */
  UINT         flag;            /* enable/disable flag */

  menu = GetMenu(hmain);        /* get menu handle */
  flag = (enable) ? MF_ENABLED : MF_GRAYED;
  EnableMenuItem(menu, MI_LOAD, flag);    /* enable/disable context- */
  EnableMenuItem(menu, MI_WGTS, flag);    /* sensitive menu items */
  mii.cbSize     = sizeof(MENUITEMINFO);  /* set either the */
  mii.fMask      = MIIM_ID|MIIM_TYPE;     /* 'Start Training' or */
  mii.fType      = MFT_STRING;            /* 'Stop Training' item */
  mii.wID        =  (enable) ? MI_START : MI_STOP;
  LoadString(hinst, (enable) ? CS_START : CS_STOP, text, 80);
  mii.dwTypeData = text; mii.cch = strlen(text);
  SetMenuItemInfo(menu, (enable) ? MI_STOP : MI_START, FALSE, &mii);
}  /* mn_enable() */

/*----------------------------------------------------------------------
  Dialog Box Procedures
----------------------------------------------------------------------*/

void dbl_set (HWND hwnd, int item, double num)
{                               /* --- set floating point number */
  sprintf(text, "%g", num);     /* format floating point number */
  edt_set(hwnd, item, text);    /* set formatted number */
}  /* dbl_set() */

/*--------------------------------------------------------------------*/

double dbl_get (HWND hwnd, int item)
{                               /* --- get floating point number */
  edt_get(hwnd, item, text, 64);    /* get input text and */
  return strtod(text, NULL);    /* convert it to a number */
}  /* dbl_get() */

/*--------------------------------------------------------------------*/

BOOL CALLBACK dp_about (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{                               /* --- 'about' dialog box procedure */
  switch (msg) {                /* evaluate message */
    case WM_COMMAND:            /* --- if command */
      if (wp != IDOK) break;    /* process only 'Ok' button */
    case WM_CLOSE:              /* --- if dialog closed */
      EndDialog(hwnd, 0);       /* terminate dialog */
      return 1;                 /* return 'ok' */
  }
  return 0;                     /* everything else not handled */
}  /* dp_about() */

/*--------------------------------------------------------------------*/

BOOL CALLBACK dp_params (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{                               /* --- parameters dialog procedure */
  double lrate, moment;         /* learning rate and momentum */

  switch (msg) {                /* evaluate message */
    case WM_INITDIALOG:         /* dialog initialization */
      LoadString(hinst, CS_ONLINE, text, 80);
      cmb_add(hwnd, DI_LMODE, text);
      LoadString(hinst, CS_BATCH,  text, 80);
      cmb_add(hwnd, DI_LMODE,  text);
      cmb_set(hwnd, DI_LMODE,  (mlp_getmode() == MLP_ONLINE) ? 0 : 1);
      dbl_set(hwnd, DI_LRATE,  mlp_getlrate());
      dbl_set(hwnd, DI_MOMENT, mlp_getmoment());
      dbl_set(hwnd, DI_RANGE,  range);
      dbl_set(hwnd, DI_DELAY,  delay);
      break;                    /* set learning parameters */
    case WM_COMMAND:            /* command */
      switch (wp) {             /* evaluate command code */
        case IDOK:              /* if 'Ok' button pressed */
          mlp_setmode((cmb_get(hwnd, DI_LMODE) == 0)
                      ? MLP_ONLINE : MLP_BATCH);
          lrate  = dbl_get(hwnd, DI_LRATE);
          if      (lrate < 1e-12) lrate  = 1e-12;
          else if (lrate > 10)    lrate  = 10;
          mlp_setlrate(lrate);  /* get the learning rate */
          moment = dbl_get(hwnd, DI_MOMENT);
          if      (moment < 0)    moment = 0;
          else if (moment > 100)  moment = 100;
          mlp_setmoment(moment);/* get the momentum */
          range  = dbl_get(hwnd, DI_RANGE);
          if      (range < 1e-12) range  = 1e-1;
          else if (range > 10)    range  = 10;
          delay  = dbl_get(hwnd, DI_DELAY);
          if      (delay < 0.001) delay  = 0.001;
          else if (delay > 10)    delay  = 10;
          EndDialog(hwnd, 0);   /* get inputs from edit controls */
          return 1;             /* return 'ok' */
        case IDCANCEL:          /* if 'Cancel' button pressed, */
          EndDialog(hwnd, -1); return 1;   /* terminate dialog */
      } break;
    case WM_CLOSE:              /* dialog box has been closed */
      EndDialog(hwnd, -1); return 1;      /* terminate dialog */
  }
  return 0;                     /* everything else not handled */
}  /* dp_params() */

/*--------------------------------------------------------------------*/

BOOL CALLBACK dp_wgts (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{                               /* --- weights dialog procedure */
  switch (msg) {                /* evaluate message */
    case WM_INITDIALOG:         /* dialog initialization */
      dbl_set(hwnd, DI_LOWER  +0,  mlp_getwgt(MLP_HD1, MLP_IN1));
      dbl_set(hwnd, DI_LOWER  +1,  mlp_getwgt(MLP_HD1, MLP_IN2));
      dbl_set(hwnd, DI_LOWER  +2, -mlp_getwgt(MLP_HD1, MLP_BIAS));
      dbl_set(hwnd, DI_UPPER  +0,  mlp_getwgt(MLP_HD2, MLP_IN1));
      dbl_set(hwnd, DI_UPPER  +1,  mlp_getwgt(MLP_HD2, MLP_IN2));
      dbl_set(hwnd, DI_UPPER  +2, -mlp_getwgt(MLP_HD2, MLP_BIAS));
      dbl_set(hwnd, DI_OUTPUT +0,  mlp_getwgt(MLP_OUT, MLP_IN1));
      dbl_set(hwnd, DI_OUTPUT +1,  mlp_getwgt(MLP_OUT, MLP_IN2));
      dbl_set(hwnd, DI_OUTPUT +2, -mlp_getwgt(MLP_OUT, MLP_BIAS));
      break;                    /* set connection weights */
    case WM_COMMAND:            /* command */
      switch (wp) {             /* evaluate command code */
        case IDOK:              /* if 'Ok' button pressed */
          mlp_setwgt(MLP_HD1, MLP_IN1,   dbl_get(hwnd, DI_LOWER  +0));
          mlp_setwgt(MLP_HD1, MLP_IN2,   dbl_get(hwnd, DI_LOWER  +1));
          mlp_setwgt(MLP_HD1, MLP_BIAS, -dbl_get(hwnd, DI_LOWER  +2));
          mlp_setwgt(MLP_HD2, MLP_IN1,   dbl_get(hwnd, DI_UPPER  +0));
          mlp_setwgt(MLP_HD2, MLP_IN2,   dbl_get(hwnd, DI_UPPER  +1));
          mlp_setwgt(MLP_HD2, MLP_BIAS, -dbl_get(hwnd, DI_UPPER  +2));
          mlp_setwgt(MLP_OUT, MLP_IN1,   dbl_get(hwnd, DI_OUTPUT +0));
          mlp_setwgt(MLP_OUT, MLP_IN2,   dbl_get(hwnd, DI_OUTPUT +1));
          mlp_setwgt(MLP_OUT, MLP_BIAS, -dbl_get(hwnd, DI_OUTPUT +2));
          EndDialog(hwnd, 0);   /* get inputs from edit controls */
          return 1;             /* return 'ok' */
        case IDCANCEL:          /* if 'Cancel' button pressed, */
          EndDialog(hwnd, -1); return 1;   /* terminate dialog */
      } break;
    case WM_CLOSE:              /* dialog box has been closed */
      EndDialog(hwnd, -1); return 1;      /* terminate dialog */
  }
  return 0;                     /* everything else not handled */
}  /* dp_wgts() */

/*----------------------------------------------------------------------
  Window Procedure
----------------------------------------------------------------------*/

LRESULT CALLBACK wp_main (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{                               /* --- main window procedure */
  static PAINTSTRUCT ps;        /* for BeginPaint call */
  HDC hdc;                      /* handle of device context */

  switch (msg) {                /* evaluate message */
    case WM_PAINT:              /* --- if to paint window */
      hdc = BeginPaint(hwnd, &ps);
      if (!hdc) return 0;       /* get device context for drawing */
      redraw(hdc);              /* redraw the window contents */
      EndPaint(hwnd, &ps);      /* and release the device context */
      break;

    case WM_COMMAND:            /* --- if menu item selected */
      switch (wp) {             /* evaluate item */
        case MI_LOAD:           /* Menu Item `Load Weights...' */
          if (load() == 0) redraw(NULL);
          return 0;             /* load weights and redraw window */
        
        case MI_SAVE:           /* Menu Item `Save Weights...' */
          save(); return 0;     /* save weights */

        case MI_QUIT:           /* Menu Item 'Quit' */
          DestroyWindow(hwnd);  /* stop network training */
          return 0;             /* and destroy window */
        
        case MI_REDRAW:         /* Menu Item 'Redraw' */
          InvalidateRect(hwnd, NULL, 0);
          return 0;             /* invalidate window contents */
        
        case MI_XOR:            /* Menu Item 'Xor' */
          init(MLP_XOR);        /* reinitialize the network */
          InvalidateRect(hwnd, NULL, 0);
          return 0;             /* invalidate window contents */

        case MI_BIIMP:          /* Menu Item 'Biimplication' */
          init(MLP_BIIMP);      /* reinitialize the network */
          InvalidateRect(hwnd, NULL, 0);
          return 0;             /* invalidate window contents */

        case MI_FUNC1:          /* Menu Item 'Function 1' */
          init(MLP_FUNC1);      /* reinitialize the network */
          InvalidateRect(hwnd, NULL, 0);
          return 0;             /* invalidate window contents */

        case MI_FUNC2:          /* Menu Item 'Function 2' */
          init(MLP_FUNC2);      /* reinitialize the network */
          InvalidateRect(hwnd, NULL, 0);
          return 0;             /* invalidate window contents */

        case MI_START:          /* Menu Item `Start Training' */
          mn_enable(0);         /* disable certain menu items */
          running = 1;          /* set the running flag */
          train(hmain, WM_TIMER, 1, 0);
          return 0;             /* start the training */

        case MI_STOP:           /* Menu Item `Stop Training' */
          KillTimer(hmain, 1);  /* kill the timer */
          mn_enable(1);         /* enable all menu items */
          return running = 0;   /* clear the running flag */
        
        case MI_PARAMS:         /* Menu Item 'Parameters...' */
          if (DialogBox(hinst, MKIRSC(DB_PARAMS), hwnd, dp_params) == 0)
            InvalidateRect(hwnd, NULL, 0);
          return 0;             /* get learning parameters */
        
        case MI_WGTS:           /* Menu Item 'Weights...' */
          if (DialogBox(hinst, MKIRSC(DB_WEIGHTS), hwnd, dp_wgts) == 0)
            InvalidateRect(hwnd, NULL, 0);
          return 0;             /* get weights of neurons */

        case MI_ABOUT:          /* Menu Item 'About' */
          DialogBox(hinst, MKIRSC(DB_ABOUT), hwnd, dp_about);
          return 0;             /* show dialog box 'About Wmlp' */
      } break;

    case WM_CLOSE:              /* window closed */
    case WM_DESTROY:            /* or destroyed */
      PostQuitMessage(0);       /* terminate program */
      return 0;
  }
  return DefWindowProc(hwnd, msg, wp, lp);
}  /* wp_main() */

/*----------------------------------------------------------------------
  Main Program
----------------------------------------------------------------------*/

int WINAPI WinMain (HINSTANCE hcurr, HINSTANCE hprev,
                    LPSTR cmdln, int cmd)
{                               /* --- main function */
  WNDCLASS wc;                  /* window class */
  RECT     rc;                  /* window rectangle */
  MSG      msg;                 /* message */

  /* --- create window class --- */
  hinst = hcurr;                /* note instance handle */
  if (!hprev) {                 /* if first instance of this program */
    wc.style         = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc   = wp_main;
    wc.cbClsExtra    = wc.cbWndExtra = 0;
    wc.hInstance     = hinst;
    wc.hIcon         = LoadIcon(NULL, IDI_APPLICATION);
    wc.hCursor       = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = NULL;
    wc.lpszMenuName  = MKIRSC(MN_MAIN);
    wc.lpszClassName = "WmlpClass";
    if (!RegisterClass(&wc))    /* register class */
      return 0;                 /* of main window */
  }

  /* --- get pens --- */
  pen_black   = GetStockObject(BLACK_PEN);
  pen_grey    = CreatePen(PS_SOLID, 1, RGB(0xc0, 0xc0, 0xc0));
  pen_red     = CreatePen(PS_SOLID, 1, RGB(0xff, 0x00, 0x00));
  pen_green   = CreatePen(PS_SOLID, 1, RGB(0x00, 0xff, 0x00));
  pen_blue    = CreatePen(PS_SOLID, 1, RGB(0x00, 0x00, 0xff));
  pen_yellow  = CreatePen(PS_SOLID, 1, RGB(0xff, 0xff, 0x00));
  pen_magenta = CreatePen(PS_SOLID, 1, RGB(0xff, 0x00, 0xff));
  pen_cyan    = CreatePen(PS_SOLID, 1, RGB(0x00, 0xc0, 0xc0));
  pen_orange  = CreatePen(PS_SOLID, 1, RGB(0xff, 0xa0, 0x00));

  /* --- get brushes --- */
  br_white    = GetStockObject(WHITE_BRUSH);
  br_black    = GetStockObject(BLACK_BRUSH);
  br_grey     = GetStockObject(LTGRAY_BRUSH);
  br_red      = CreateSolidBrush(RGB(0xff, 0x00, 0x00));
  br_green    = CreateSolidBrush(RGB(0x00, 0xff, 0x00));
  br_blue     = CreateSolidBrush(RGB(0x00, 0x00, 0xff));
  br_yellow   = CreateSolidBrush(RGB(0xff, 0xff, 0x00));
  br_magenta  = CreateSolidBrush(RGB(0xff, 0x00, 0xff));
  br_cyan     = CreateSolidBrush(RGB(0x00, 0xc0, 0xc0));
  br_orange   = CreateSolidBrush(RGB(0xff, 0xa0, 0x00));

  srand((unsigned long)time(NULL));
  init(MLP_XOR);                /* initialize the neural network */

  /* --- create window --- */
  rc.left = 0; rc.right  = INITXEXT -1;
  rc.top  = 0; rc.bottom = INITYEXT -1;
  AdjustWindowRect(&rc, WINSTYLE, 1);
  hmain = CreateWindow("WmlpClass", PRGNAME, WINSTYLE,
                       30, 30, rc.right-rc.left, rc.bottom-rc.top,
                       NULL, NULL, hinst, NULL);
  if (!hmain) return 0;         /* create main window */
  hdc_win = GetDC(hmain);
  if (!hdc_win) error(E_INIT);
  hdc_buf = CreateCompatibleDC(hdc_win);
  if (!hdc_buf) error(E_INIT);
  buffer  = CreateCompatibleBitmap(hdc_win, INITXEXT, INITYEXT);
  if (!buffer) error(E_INIT);
  SelectObject(hdc_buf, buffer);
  SelectObject(hdc_buf, GetStockObject(ANSI_VAR_FONT));
  xoff = 0; yoff = INITYEXT -1; /* get coordinate offsets */
  ShowWindow(hmain, cmd);       /* open and */
  UpdateWindow(hmain);          /* update main window */

  /* --- dispatch messages --- */
  while (GetMessage(&msg, NULL, 0, 0) > 0) {
    TranslateMessage(&msg);     /* get messages, translate */
    DispatchMessage (&msg);     /* and dispatch them */
  }

  /* --- clean up --- */
  ReleaseDC(hmain, hdc_win);    /* release device context of window */
  DeleteObject(hdc_buf);        /* delete device context of buffer */
  DeleteObject(buffer);         /* delete drawing buffer */
  DeleteObject(br_red);         /* delete created brushs */
  DeleteObject(br_green);
  DeleteObject(br_blue);
  DeleteObject(br_yellow);
  DeleteObject(br_grey);
  DeleteObject(br_magenta);
  DeleteObject(br_cyan);
  DeleteObject(br_orange);
  DeleteObject(pen_red);        /* delete created pens */
  DeleteObject(pen_green);
  DeleteObject(pen_blue);
  DeleteObject(pen_yellow);
  DeleteObject(pen_grey);
  DeleteObject(pen_magenta);
  DeleteObject(pen_cyan);
  DeleteObject(pen_orange);
  return msg.wParam;
}  /* WinMain() */
