/*----------------------------------------------------------------------
  File    : xmlp.c
  Contents: X11 program to visualize multilayer perceptron training
  Author  : Christian Borgelt
  History : 03.11.1998 file created from file xskel.c
            03.04.1999 dialog box for connection weights added
            12.01.2000 adapted to module dialog
            18.03.2000 adapted to module mlp, function approx. added
            27.08.2000 renamed to xmlp.c
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <float.h>
#include <math.h>
#include <time.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Core.h>
#include <X11/Shell.h>
#include <X11/Xaw/Form.h>
#include "menu.h"
#include "dialog.h"
#include "fselect.h"
#include "mlpd.h"

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define	M_PI	 3.14159265358979323846  /* \pi */

/* --- extensions --- */
#define INITXEXT     572        /* initial x-extension of window */
#define INITYEXT     400        /* initial y-extension of window */
#define MINVAL       (-0.2)     /* range of values */
#define MAXVAL         1.2      /* for diagrams */
#define FMT_MAX      256        /* max. len. of error message format */

/* --- error codes --- */
#define OK             0        /* no error */
#define E_NONE         0        /* no error */
#define E_NOMEM      (-1)       /* not enough memory */
#define E_FOPEN      (-2)       /* cannot open file */
#define E_FREAD      (-3)       /* read error on file */
#define E_FWRITE     (-4)       /* write error on file */
#define E_INIT       (-5)       /* initialization failed */
#define E_WIDGET     (-6)       /* widget creation failed */
#define E_COLOR      (-7)       /* color allocation failed */
#define E_FONT       (-8)       /* font loading failed */
#define E_GRAPH      (-9)       /* cannot create graphics context */
#define E_DIALOG    (-10)       /* cannot create dialog box */
#define E_UNKNOWN   (-11)       /* unkown error */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- error message data --- */
  const char *name, *type;      /* name and type of the error message */
  const char *class;            /* resource class of error message */
  const char *dflt;             /* default message */
  int        argc;              /* number of arguments */
} ERRMSG;                       /* (error message) */

/*----------------------------------------------------------------------
  Function Prototypes
----------------------------------------------------------------------*/
static void load    (Widget w, XEvent *e, String *s, Cardinal *c);
static void save    (Widget w, XEvent *e, String *s, Cardinal *c);
static void quit    (Widget w, XEvent *e, String *s, Cardinal *c);
static void init    (Widget w, XEvent *e, String *s, Cardinal *c);
static void params  (Widget w, XEvent *e, String *s, Cardinal *c);
static void weights (Widget w, XEvent *e, String *s, Cardinal *c);
static void exec    (Widget w, XEvent *e, String *s, Cardinal *c);
static void redraw  (Widget w, XEvent *e, String *s, Cardinal *c);
static void about   (Widget w, XEvent *e, String *s, Cardinal *c);

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
#include "xmlp.rsc"             /* fallback resources */

static Cardinal one = 1;        /* the number one (action parameter) */

static XtActionsRec actions[] = {
  { "load",     load     },     /* File    > Load File... */
  { "save",     save     },     /* File    > Save File... */
  { "quit",     quit     },     /* File    > Quit */
  { "init",     init     },     /* Actions > Xor/Biimp */
  { "params",   params   },     /* Actions > Parameters... */
  { "weights",  weights  },     /* Actions > Weights... */
  { "exec",     exec     },     /* Actions > Start/Stop */
  { "redraw",   redraw   },     /* Actions > Redraw */
  { "about",    about    },     /* Help    > About Xmlp... */
  { "db_close", db_close } };   /* close dialog box */

/* --- error messages --- */
static const ERRMSG errmsgs[] = {   /* error message data */
  /* E_NONE      0 */  { "error",   "none",    "Xmlp.Error",
                         "no error",                          0 },
  /* E_NOMEM    -1 */  { "error",   "nomem",   "Xmlp.Error",
                         "not enough memory",                 0 },
  /* E_FOPEN    -2 */  { "warning", "fopen",   "Xmlp.Warning",
                         "cannot open file:\n%s",             1 },
  /* E_FREAD    -3 */  { "warning", "fread",   "Xmlp.Warning",
                         "read error on file:\n%s",           1 },
  /* E_FWRITE   -4 */  { "warning", "fwrite",  "Xmlp.Warning",
                         "write error on file:\n%s",          1 },
  /* E_INIT     -5 */  { "error",   "init",    "Xmlp.Error",
                         "initialization failed",             0 },
  /* E_WIDGET   -6 */  { "error",   "widget",  "Xmlp.Error",
                         "widget creation failed",            0 },
  /* E_COLOR    -7 */  { "error",   "color",   "Xmlp.Error",
                         "color allocation failed",           0 },
  /* E_FONT     -8 */  { "error",   "font",    "Xmlp.Error",
                         "font loading failed",               0 },
  /* E_GRAPH    -9 */  { "error",   "gc",      "Xmlp.Error",
                         "cannot create graphics context",    0 },
  /* E_DIALOG  -10 */  { "warning", "dialog",  "Xmlp.Warning",
                         "cannot create dialog box",          0 },
  /* E_UNKNOWN -11 */  { "error", "unknown",   "Xmlp.Error",
                         "unknown error",                     0 } };

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static XtAppContext appctx;     /* X11 application context */
static Display *display;        /* display connected to */
static int     screen;          /* screen of the display */
static Window  window;          /* main window */
static Atom    wm_delwin;       /* delete window atom of window mgr. */
static char    fmt[FMT_MAX];    /* format and buffer for messages */
static char    buf[PATH_MAX +FMT_MAX];

/* --- main widgets and menu --- */
static Widget  w_top;           /* top level widget */
static Widget  w_main;          /* main window widget */
static Widget  w_view;          /* viewport widget */
static MENU    *menu;           /* pull down menu */
static GC      gc;              /* graphics context */

/* --- property sheets --- */
static PSHEET  *psh_params = NULL;  /* general parameters */
static PSHEET  *psh_wgts   = NULL;  /* connection weights */

/* --- color variables --- */
static unsigned long white;     /* white   (background) */
static unsigned long black;     /* black   (coordinate axes) */
static unsigned long grey;      /* grey    (auxiliary lines) */
static unsigned long magenta;   /* magenta (separating line) */
static unsigned long cyan;      /* cyan    (separating line) */
static unsigned long orange;    /* orange  (separating line) */
static unsigned long yellow;    /* yellow  (the point (0,0)) */
static unsigned long green;     /* green   (the point (1,0)) */
static unsigned long blue;      /* blue    (the point (0,1)) */
static unsigned long red;       /* red     (the point (1,1)) */

/* --- font variables --- */
static XFontStruct *font;       /* font information */

/* --- neural network variables --- */
static int    xoff, yoff;       /* offsets for drawing */
static double range   = 1;      /* initial weight range */
static double delay   = 0.01;   /* delay between training steps */
static int    diag    = 0;      /* whether diagram has been drawn */
static int    running = 0;      /* whether training is running */
static Pixmap buffer;           /* buffer for diagram drawing */
static XtIntervalId iid;        /* id of time interval procedure */

/*----------------------------------------------------------------------
  Auxiliary Functions
----------------------------------------------------------------------*/

static void dcb_nop (Widget widget, XtPointer client, XtPointer call)
{                               /* --- close dialog box (callback) */
  while (XtClass(widget) != transientShellWidgetClass)
    widget = XtParent(widget);  /* find the shell widget */
  XtPopdown(widget);            /* close dialog box */
  mn_enable(menu, MN_MENU, 1);  /* and enable menu bar */
}  /* dcb_nop() */

/*--------------------------------------------------------------------*/

static void error (int code, ...)
{                               /* --- print error message */
  va_list      va;              /* list of variable arguments */
  const ERRMSG *msg;            /* error message */
  String       argv[8];         /* variable argument vector */
  Cardinal     argc;            /* number of arguments */

  if ((code > 0) || (code < E_UNKNOWN))
    code = E_UNKNOWN;           /* check error code */
  msg = errmsgs -code;          /* get error message data and */
  va_start(va, code);           /* list of variable arguments */
  for (argc = 0; (int)argc < msg->argc; argc++)
    argv[argc] = va_arg(va, String);
  va_end(va);                   /* get message arguments */
  if (strcmp(msg->name, "error") == 0)  /* if name is ``error'' */
    XtAppErrorMsg(appctx, msg->name, msg->type, msg->class,
      msg->dflt, argv, &argc);  /* print error message */
  else {                        /* if name is ``warning'' */
    XtAppGetErrorDatabaseText(appctx, msg->name, msg->type,
      msg->class, msg->dflt, fmt, FMT_MAX, NULL);
    va_start(va, code);         /* list of variable arguments */
    vsprintf(buf, fmt, va);     /* get and format message */
    va_end(va);                 /* end variable argument evaluation */
    mn_enable(menu, MN_MENU,0); /* disable menu bar and show dialog */
    if (db_alert(w_top, dcb_nop, buf) != 0) {
      mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
      XtAppWarningMsg(appctx, msg->name, msg->type, msg->class,
        msg->dflt, argv, &argc);/* if the alert dialog box */
    }                           /* cannot be shown, */
  }                             /* print a warning message */
}  /* error() */

/*----------------------------------------------------------------------
  Drawing Functions
----------------------------------------------------------------------*/

static void arrow (int x1, int y1, int x2, int y2)
{                               /* --- draw a horz. or vert. arrow */
  if (x1 == x2) {               /* if vertical arrow */
    XDrawLine(display, buffer, gc, x1, y1, x2, y2+6);
    XFillArc (display, buffer, gc, x2-8, y2-8, 16, 16, 250*64, 40*64); }
  else {                        /* if horizontal arrow */
    XDrawLine(display, buffer, gc, x1, y1, x2-6, y2);
    XFillArc (display, buffer, gc, x2-8, y2-8, 16, 16, 160*64, 40*64);
  }
}  /* arrow() */

/*--------------------------------------------------------------------*/

static void conwgt (int x, int y, double wgt)
{                               /* --- print connection weight */
  int  len, wd;                 /* string length and text width */
  char s[32];                   /* buffer for weight as a text */
 
  len = sprintf(s," %.4g ",wgt);/* convert the weight to a string */
  wd  = XTextWidth(font, s,len);/* and get its length and width */
  XDrawImageString(display, buffer, gc, x -wd/2, y, s, len);
}  /* conwgt() */               /* draw the connection weight */

/*--------------------------------------------------------------------*/

static void point (double x, double y, int value,
                   int diag, int size, unsigned long color)
{                               /* --- draw a (transformed) point */
  int wx = xoff +(diag > 0) ? 328 : 44;
  int wy = yoff -44;            /* window coordinates */

  wx += (int)(x *200) -size/2;  /* compute window coordinates */
  wy -= (int)(y *200) +size/2;  /* and draw point */
  XSetForeground(display, gc, (value) ? color : white);
  XFillArc(display, buffer, gc, wx+1, wy+1, size-2, size-2, 0, 360*64);
  XSetForeground(display, gc, color);
  XDrawArc(display, buffer, gc, wx+1, wy+1, size-2, size-2, 0, 360*64);
  XDrawArc(display, buffer, gc, wx,   wy,   size,   size,   0, 360*64);
}  /* point() */

/*--------------------------------------------------------------------*/

static void line (double a, double b, double c, int diag,
                  unsigned long color)
{                               /* --- draw a separating line */
  int    wx = xoff +(diag > 0) ? 328 : 44;
  int    wy = yoff -44;         /* window coordinates */
  double x1, y1, x2, y2;        /* coordinates for lines */
  double t;                     /* temporary buffer */

  /* --- clipping --- */
  if (fabs(a) < 1e-12) a = (a < 0) ? -1e-12 : 1e-12;
  if (fabs(b) < 1e-12) b = (b < 0) ? -1e-12 : 1e-12;
  x1 = (-MINVAL *b -c) /a;
  if      (x1 < MINVAL) { x1 = MINVAL; y1 = (-MINVAL *a -c) /b; }
  else if (x1 > MAXVAL) { x1 = MAXVAL; y1 = (-MAXVAL *a -c) /b; }
  else                  {              y1 = MINVAL; }
  x2 = (-MAXVAL *b -c) /a;
  if      (x2 < MINVAL) { x2 = MINVAL; y2 = (-MINVAL *a -c) /b; }
  else if (x2 > MAXVAL) { x2 = MAXVAL; y2 = (-MAXVAL *a -c) /b; }
  else                  {              y2 = MAXVAL; }
  if ((y1 < MINVAL) || (y1 > MAXVAL)
  ||  (y2 < MINVAL) || (y2 > MAXVAL))
    return;                     /* skip invisible lines */

  /* --- drawing --- */
  XSetForeground(display, gc, color);
  XDrawLine(display, buffer, gc,
            wx +(int)(x1 *200), wy -(int)(y1 *200),
            wx +(int)(x2 *200), wy -(int)(y2 *200));
  t  = sqrt(a*a + b*b);
  x1 = (x1 +x2) /2; x2 = x1 +0.05 *a/t;
  y1 = (y1 +y2) /2; y2 = y1 +0.05 *b/t;
  XDrawLine(display, buffer, gc,
            wx +(int)(x1 *200), wy -(int)(y1 *200),
            wx +(int)(x2 *200), wy -(int)(y2 *200));
  XFillArc (display, buffer, gc,
            wx +(int)(x2 *200) -8, wy -(int)(y2 *200)-8, 16, 16,
            (atan2(y1 -y2, x1 -x2) *(180/M_PI) -20)*64, 40*64);
}  /* line() */

/*--------------------------------------------------------------------*/

static void patterns (void)
{                               /* --- draw training patterns */
  int cnt;                      /* number of patterns */
  int x0 = xoff +INITXEXT/2;    /* x-coordinate of origin */
  int y0 = yoff -28;            /* y-coordinate of origin */
  const PAT *p;                 /* to traverse the patterns */
  
  XSetForeground(display, gc, black);
  cnt = mlp_patcnt();           /* get the number of patterns */
  while (--cnt >= 0) {          /* and traverse the patterns */
    p = mlp_pattern(cnt);       /* get the next training pattern */
    XFillRectangle(display, buffer, gc, x0 +(int)(230 *p->in1)-1,
                                        y0 -(int)(230 *p->out)-1, 3, 3);
  }                             /* draw circles at the data points */
}  /* patterns() */

/*--------------------------------------------------------------------*/

static void netout (void)
{                               /* --- draw network output */
  static int hd1[101];          /* hidden neuron 1 output */
  static int hd2[101];          /* hidden neuron 2 output */
  static int out[101];          /* output neuron output */
  int    x0 = xoff +36;         /* x-coordinate of reference point */
  int    y0 = yoff -28;         /* y-coordinate of reference point */
  int    x;                     /* loop variable, pixel coordinate */

  for (x = 100; --x >= 0; ) {   /* traverse the input range */
    out[x] = y0 -(int)(230 *mlp_calc((x-50)*(5/230.0), 0));
    hd1[x] = y0 -(int)(230 *mlp_output(0));
    hd2[x] = y0 -(int)(230 *mlp_output(1));
  }                             /* compute outputs of neurons */
  XSetForeground(display, gc, blue);
  for (x = 100; --x > 0; )
    XDrawLine(display, buffer, gc, x0+5*x-5, hd1[x-1], x0+5*x, hd1[x]);
  XSetForeground(display, gc, green);
  for (x = 100; --x > 0; )
    XDrawLine(display, buffer, gc, x0+5*x-5, hd2[x-1], x0+5*x, hd2[x]);
  XSetForeground(display, gc, red);
  for (x = 100; --x > 0; )
    XDrawLine(display, buffer, gc, x0+5*x-5, out[x-1], x0+5*x, out[x]);
}  /* netout() */

/*--------------------------------------------------------------------*/

static void draw_all (void)
{                               /* --- draw the window contents */
  int type;                     /* problem type */
  int bool;                     /* Boolean function flag */
  const PAT *p;                 /* to traverse the training patterns */

  type = mlp_gettype();         /* get the problem type and */
  bool = (type < MLP_FUNC1);    /* compute Boolean function flag */
  if (diag) {                   /* if diagram has been drawn once */
    XSetForeground(display, gc, white);
    XFillRectangle(display, buffer, gc, 0, yoff-294, INITXEXT, 294);
    XFillRectangle(display, buffer, gc, xoff+126, yoff-390, 138, 93);
    XFillRectangle(display, buffer, gc, xoff+301, yoff-390, 137, 93);
    XFillRectangle(display, buffer, gc, xoff+438, yoff-331,  92, 34);
    XFillRectangle(display, buffer, gc,
                   xoff+ ((bool) ? 86 : 264),     yoff-366,  37, 45); }
  else {                        /* if diagram has not been drawn yet */
    XSetForeground(display, gc, white);
    XFillRectangle(display, buffer, gc, 0, 0, INITXEXT, INITYEXT);

    /* --- neurons --- */
    XSetForeground(display, gc, grey);
    if (!bool) {                /* one input for real-valued function */
      XFillArc(display, buffer, gc, xoff+100,yoff-355, 22,22,0,360*64);
      XSetForeground(display, gc, black);
      XDrawArc(display, buffer, gc, xoff+100,yoff-355, 22,22,0,360*64);}
    else {                      /* two inputs for Boolean function */
      XFillArc(display, buffer, gc, xoff+100,yoff-320, 22,22,0,360*64);
      XFillArc(display, buffer, gc, xoff+100,yoff-390, 22,22,0,360*64);
      XSetForeground(display, gc, black);
      XDrawArc(display, buffer, gc, xoff+100,yoff-390, 22,22,0,360*64);
      XDrawArc(display, buffer, gc, xoff+100,yoff-320, 22,22,0,360*64);
    }
    XSetForeground(display, gc, (bool) ? cyan    : blue);
    XFillArc(display, buffer, gc, xoff+275, yoff-320, 22,22, 0,360*64);
    XSetForeground(display, gc, (bool) ? magenta : green);
    XFillArc(display, buffer, gc, xoff+275, yoff-390, 22,22, 0,360*64);
    XSetForeground(display, gc, (bool) ? orange  : red);
    XFillArc(display, buffer, gc, xoff+450, yoff-355, 22,22, 0,360*64);
    XSetForeground(display, gc, black);
    XDrawArc(display, buffer, gc, xoff+275, yoff-390, 22,22, 0,360*64);
    XDrawArc(display, buffer, gc, xoff+275, yoff-320, 22,22, 0,360*64);

    /* --- connections --- */
    if (bool) {                 /* two inputs for Boolean functions */
      arrow(xoff +40, yoff-379, xoff +97, yoff-379);  /* upper */
      arrow(xoff+111, yoff-372, xoff+111, yoff-387);  /* input */
      arrow(xoff +40, yoff-309, xoff +97, yoff-309);  /* lower */
      arrow(xoff+103, yoff-309, xoff+119, yoff-309);  /* input */
      XFillArc(display, buffer, gc,     /* upper hidden neuron */
               xoff+266, yoff-380, 16,16,180*64,40*64);
      arrow(xoff+126, yoff-379, xoff+272, yoff-379);
      arrow(xoff+286, yoff-372, xoff+286, yoff-387);
      XFillArc(display, buffer, gc,     /* lower hidden neuron */
               xoff+266, yoff-324, 16,16,140*64,42*64);
      arrow(xoff+126, yoff-309, xoff+272, yoff-309);
      arrow(xoff+278, yoff-309, xoff+294, yoff-309); }
    else {                      /* one input for real-valued function */
      arrow(xoff +40, yoff-344, xoff +97, yoff-344);
      XFillArc(display, buffer, gc,
               xoff+263, yoff-387, 16,16,172*64,42*64);
      XFillArc(display, buffer, gc,
               xoff+263, yoff-317, 16,16,146*64,42*64);
    }
    XFillArc(display, buffer, gc,       /* output neuron */
             xoff+438, yoff-356, 16,16,146*64,42*64);
    XFillArc(display, buffer, gc,
             xoff+438, yoff-348, 16,16,172*64,42*64);
    XDrawArc(display, buffer, gc,
             xoff+450, yoff-355, 22, 22, 0, 360*64);
    arrow(xoff+476, yoff-344, xoff+532, yoff-344);
    diag = 1;                   /* set diagram initialization flag */
  }

  /* --- connections --- */
  XSetForeground(display, gc, black);
  if (bool) {                   /* if Boolean function */
    XDrawLine(display, buffer, gc,  /* upper hidden */
              xoff+124, yoff-317, xoff+272, yoff-371);
    XDrawLine(display, buffer, gc,
              xoff+126, yoff-379, xoff+264, yoff-379);
    XDrawLine(display, buffer, gc,  /* lower hidden */
              xoff+126, yoff-309, xoff+264, yoff-309);
    XDrawLine(display, buffer, gc,
              xoff+124, yoff-371, xoff+272, yoff-317); }
  else {                        /* if real-valued function */
    XDrawLine(display, buffer, gc,
              xoff+126, yoff-340, xoff+271, yoff-309);
    XDrawLine(display, buffer, gc,
              xoff+126, yoff-348, xoff+271, yoff-379);
  }                             /* (connections to hidden neurons) */
  XDrawLine(display, buffer, gc,
            xoff+301, yoff-379, xoff+446, yoff-348);
  XDrawLine(display, buffer, gc,
            xoff+301, yoff-309, xoff+446, yoff-340);

  /* -- connection weights -- */
  if (bool) {                   /* if Boolean function */
    conwgt(xoff+198, yoff-304, mlp_getwgt(MLP_HD1, MLP_IN1));
    conwgt(xoff+158, yoff-353, mlp_getwgt(MLP_HD1, MLP_IN2));
    conwgt(xoff+158, yoff-325, mlp_getwgt(MLP_HD2, MLP_IN1));
    conwgt(xoff+198, yoff-374, mlp_getwgt(MLP_HD2, MLP_IN2)); }
  else {                        /* if real-valued function */
    conwgt(xoff+198, yoff-319, mlp_getwgt(MLP_HD1, MLP_IN1));
    conwgt(xoff+198, yoff-359, mlp_getwgt(MLP_HD2, MLP_IN1));
  }                             /* (connections to hidden neurons) */
  conwgt(xoff+286, yoff-325, -mlp_getwgt(MLP_HD1, MLP_BIAS));
  conwgt(xoff+286, yoff-353, -mlp_getwgt(MLP_HD2, MLP_BIAS));
  conwgt(xoff+373, yoff-319,  mlp_getwgt(MLP_OUT, MLP_IN1));
  conwgt(xoff+373, yoff-359,  mlp_getwgt(MLP_OUT, MLP_IN2));
  conwgt(xoff+461, yoff-318, -mlp_getwgt(MLP_OUT, MLP_BIAS));

  /* --- separating line --- */
  XSetForeground(display, gc, grey);
  XDrawLine(display, buffer, gc, xoff, yoff-289, xoff+571, yoff-289);

  if (bool) {                   /* if Boolean function */
    /* --- left diagram --- */
    XSetForeground(display, gc, grey);
    XDrawLine(display, buffer, gc,
              xoff+244, yoff -44, xoff+244, yoff-244);
    XDrawLine(display, buffer, gc,
              xoff+ 44, yoff-244, xoff+244, yoff-244);
    XSetForeground(display, gc, black);
    arrow(xoff +44, yoff -44, xoff +44, yoff-264);
    arrow(xoff +44, yoff -44, xoff+265, yoff -44);
    p = mlp_pattern(0); point(p->in1, p->in2, (int)p->out, 0, 8, yellow);
    p = mlp_pattern(1); point(p->in1, p->in2, (int)p->out, 0, 8, green);
    p = mlp_pattern(2); point(p->in1, p->in2, (int)p->out, 0, 8, blue);
    p = mlp_pattern(3); point(p->in1, p->in2, (int)p->out, 0, 8, red);
    line(mlp_getwgt(MLP_HD1, MLP_IN1), mlp_getwgt(MLP_HD1, MLP_IN2),
         mlp_getwgt(MLP_HD1, MLP_BIAS), 0, cyan);
    line(mlp_getwgt(MLP_HD2, MLP_IN1), mlp_getwgt(MLP_HD2, MLP_IN2),
         mlp_getwgt(MLP_HD2, MLP_BIAS), 0, magenta);

    /* --- right diagram --- */
    XSetForeground(display, gc, grey);
    XDrawLine(display, buffer, gc,
              xoff+428, yoff -44, xoff+428, yoff-244);
    XDrawLine(display, buffer, gc,
              xoff+328, yoff-144, xoff+528, yoff-144);
    XDrawLine(display, buffer, gc,
              xoff+528, yoff -44, xoff+528, yoff-244);
    XDrawLine(display, buffer, gc,
              xoff+328, yoff-244, xoff+528, yoff-244);
    XSetForeground(display, gc, black);
    arrow(xoff+328, yoff -44, xoff+328, yoff-264);
    arrow(xoff+328, yoff -44, xoff+549, yoff -44);
    p = mlp_pattern(0); mlp_calc(p->in1, p->in2);
    point(mlp_output(0), mlp_output(1), (int)p->out, 1, 8, yellow);
    p = mlp_pattern(1); mlp_calc(p->in1, p->in2);
    point(mlp_output(0), mlp_output(1), (int)p->out, 1, 8, green);
    p = mlp_pattern(2); mlp_calc(p->in1, (int)p->in2);
    point(mlp_output(0), mlp_output(1), (int)p->out, 1, 8, blue);
    p = mlp_pattern(3); mlp_calc(p->in1, (int)p->in2);
    point(mlp_output(0), mlp_output(1), (int)p->out, 1, 8, red);
    line(mlp_getwgt(MLP_OUT, MLP_IN1), mlp_getwgt(MLP_OUT, MLP_IN2),
         mlp_getwgt(MLP_OUT, MLP_BIAS), 1, orange); }

  else {                        /* if real-valued function */
    /* --- diagram --- */
    XSetForeground(display, gc, grey);
    XDrawLine(display, buffer, gc,
              xoff +20, yoff-258, xoff+552, yoff-258);
    XDrawLine(display, buffer, gc,
              xoff +20, yoff-143, xoff+552, yoff-143);
    XSetForeground(display, gc, black);
    arrow(xoff+286, yoff -25, xoff+286, yoff-274);
    arrow(xoff+ 21, yoff -28, xoff+552, yoff -28);
    XFillArc (display, buffer, gc,
              xoff +12, yoff -36, 16,16,-20*64,40*64);
    XDrawLine(display, buffer, gc,
              xoff+283, yoff-258, xoff+289, yoff-258);
    XDrawLine(display, buffer, gc,
              xoff +56, yoff -25, xoff +56, yoff -31);
    XDrawLine(display, buffer, gc,
              xoff+516, yoff -25, xoff+516, yoff -31);
    patterns();                 /* draw the training data and */
    netout();                   /* the outputs of the neurons */
  }
}  /* draw_all() */

/*----------------------------------------------------------------------
  Dialog Box Callback Functions
----------------------------------------------------------------------*/

static void dcb_load (Widget widget, XtPointer ok, XtPointer fname)
{                               /* --- 'Load Weights...' callback */
  FILE   *file;                 /* input file */
  int    i;                     /* loop variable */
  double w[3][3];               /* buffer for weights */

  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
  if (!ok) return;              /* if cancel button pressed, abort */
  file = fopen(fname, "r");     /* open input file */
  if (!file) { error(E_FOPEN, fname); return; }
  for (i = 0; i < 3; i++) {     /* read weights from file */
    if (fscanf(file, "%lf%lf%lf", w[i], w[i] +1, w[i] +2) != 3) {
      error(E_FREAD, fname); return; }
  }                             /* store weights in buffer */
  fclose(file);                 /* close input file */
  for (i = 3; --i >= 0; ) {     /* set the connection weights */
    mlp_setwgt(i, MLP_IN1,   w[i][0]);     /* from the buffer */
    mlp_setwgt(i, MLP_IN2,   w[i][1]);
    mlp_setwgt(i, MLP_BIAS, -w[i][2]);
  }
  mlp_init(-DBL_MAX);           /* initialize the neural network */
  redraw(NULL,NULL,NULL,NULL);  /* and redraw the window contents */
}  /* dcb_load() */

/*--------------------------------------------------------------------*/

static void dcb_save (Widget widget, XtPointer ok, XtPointer fname)
{                               /* --- 'Save Weights...' callback */
  FILE *file;                   /* output file */
  int  i;                       /* loop variable */

  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
  if (!ok) return;              /* if cancel button pressed, abort */
  file = fopen(fname, "w");     /* open output file */
  if (!file) { error(E_FOPEN, fname); return; }
  for (i = 0; i < 3; i++)       /* write the connection weights */
    fprintf(file, "%g %g %g\n", /* to the chosen file */
            mlp_getwgt(i, MLP_IN1),
            mlp_getwgt(i, MLP_IN2),
           -mlp_getwgt(i, MLP_BIAS));
  if (fclose(file) != 0) error(E_FWRITE, fname);
}  /* dcb_save() */

/*--------------------------------------------------------------------*/

static void dcb_params (Widget widget, XtPointer ok, XtPointer call)
{                               /* --- 'Parameters...' callback */
  double lrate, moment;         /* learning rate and momentum */

  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
  if (!ok) return;              /* if not ok pressed, abort */
  mlp_setmode(((int)psh_getval(psh_params, "mode") == 0)
              ? MLP_ONLINE : MLP_BATCH);
  lrate  = atof(psh_getval(psh_params, "lrate"));
  if      (lrate < 1e-12) lrate  = 1e-12;
  else if (lrate > 10)    lrate  = 10;
  mlp_setlrate(lrate);          /* get the learning rate */
  moment = atof(psh_getval(psh_params, "moment"));
  if      (moment < 0)    moment = 0;
  else if (moment > 10)   moment = 10;
  mlp_setmoment(moment);        /* get the momentum */
  range  = atof(psh_getval(psh_params, "range"));
  if      (range < 1e-12) range  = 1e-1;   /* get the initial */
  else if (range > 10)    range  = 10;     /* weight range */
  delay  = atof(psh_getval(psh_params, "delay"));
  if      (delay < 0.001) delay  = 0.001;  /* get the delay */
  else if (delay > 10)    delay  = 10;     /* between steps */
  redraw(NULL,NULL,NULL,NULL);  /* redraw the window contents */
}  /* dcb_params() */

/*--------------------------------------------------------------------*/

static void dcb_wgts (Widget widget, XtPointer ok, XtPointer call)
{                               /* --- 'Weights...' callback */
  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
  if (!ok) return;              /* if not ok pressed, abort */
  mlp_setwgt(0, 0,  atof(psh_getval(psh_wgts, "lower_0")));
  mlp_setwgt(0, 1,  atof(psh_getval(psh_wgts, "lower_1")));
  mlp_setwgt(0, 2, -atof(psh_getval(psh_wgts, "lower_2")));
  mlp_setwgt(1, 0,  atof(psh_getval(psh_wgts, "upper_0")));
  mlp_setwgt(1, 1,  atof(psh_getval(psh_wgts, "upper_1")));
  mlp_setwgt(1, 2, -atof(psh_getval(psh_wgts, "upper_2")));
  mlp_setwgt(2, 0,  atof(psh_getval(psh_wgts, "output_0")));
  mlp_setwgt(2, 1,  atof(psh_getval(psh_wgts, "output_1")));
  mlp_setwgt(2, 2, -atof(psh_getval(psh_wgts, "output_2")));
  redraw(NULL,NULL,NULL,NULL);  /* get the new weights and */
}  /* dcb_weights() */          /* redraw the window contents */

/*----------------------------------------------------------------------
  Menu Callback Functions
----------------------------------------------------------------------*/

static void mcb_load   (Widget widget, XtPointer client, XtPointer call)
{ load(NULL, NULL, NULL, NULL); }

static void mcb_save   (Widget widget, XtPointer client, XtPointer call)
{ save(NULL, NULL, NULL, NULL); }

static void mcb_quit   (Widget widget, XtPointer client, XtPointer call)
{ quit(NULL, NULL, NULL, NULL); }

static void mcb_init   (Widget widget, XtPointer client, XtPointer call)
{ init(NULL, NULL, (String*)&client, &one); }

static void mcb_params (Widget widget, XtPointer client, XtPointer call)
{ params(NULL, NULL, NULL, NULL); }

static void mcb_wgts   (Widget widget, XtPointer client, XtPointer call)
{ weights(NULL, NULL, NULL, NULL); }

static void mcb_start  (Widget widget, XtPointer client, XtPointer call)
{ exec(NULL, NULL, NULL, NULL); }

static void mcb_redraw (Widget widget, XtPointer client, XtPointer call)
{ diag = 0; redraw(NULL, NULL, NULL, NULL); }

static void mcb_about  (Widget widget, XtPointer client, XtPointer call)
{ about(NULL, NULL, NULL, NULL); }

/*----------------------------------------------------------------------
  Neural Network Training Function
----------------------------------------------------------------------*/

static void train (XtPointer data, XtIntervalId *id)
{                               /* --- do one training step */
  if (!running) return;         /* check running flag */
  mlp_train();                  /* train the neural network and */
  redraw(NULL,NULL,NULL,NULL);  /* redraw the window contents */
  iid = XtAppAddTimeOut(appctx, (unsigned long)(delay *1000),
          train, NULL);         /* reregister interval procedure */
}  /* train() */

/*----------------------------------------------------------------------
  Action Functions
----------------------------------------------------------------------*/

static void load (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- File > Load... */
  const char *t;                /* buffer for dialog box title */

  if (running) return;          /* if training is running, abort */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  XtVaGetValues(XtNameToWidget(mn_menubar(menu), "*load"),
    XtNlabel, &t, NULL);        /* get name of menu entry */
  if (fselect(w_top, t, NULL, NULL, dcb_load) != 0) {
    error(E_DIALOG); mn_enable(menu, MN_MENU, 1); }
}  /* load() */

/*--------------------------------------------------------------------*/

static void save (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- File > Save... */
  const char *t;                /* buffer for dialog box title */

  if (running) return;          /* if training is running, abort */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  XtVaGetValues(XtNameToWidget(mn_menubar(menu), "*save"),
    XtNlabel, &t, NULL);        /* get name of menu entry */
  if (fselect(w_top, t, NULL, NULL, dcb_save) != 0) {
    error(E_DIALOG); mn_enable(menu, MN_MENU, 1); }
}  /* save() */

/*--------------------------------------------------------------------*/

static void quit (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- File > Quit */
  if (w_top) XtDestroyApplicationContext(appctx);
  exit(0);                      /* terminate program */
}  /* quit() */

/*--------------------------------------------------------------------*/

static void redraw (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Redraw */
  XRectangle rc;                /* clipping rectangle */
  int        i;                 /* loop variable */
  char       buf[32];           /* text buffer for epoch */

  if (!gc) {                    /* if graphics context not created */
    gc = XCreateGC(display, window, 0, NULL);
    if (!gc) { error(E_GRAPH); return; } /* create a graphics context */
    i = XDefaultDepth(display, screen);
    buffer = XCreatePixmap(display, window, INITXEXT, INITYEXT, i);
    if (buffer == None) {       /* create a buffer for drawing */
      XFreeGC(display, gc); gc = NULL; error(E_GRAPH); return; }
    XSetFont(display, gc, font->fid);   /* set the font */
    XSetBackground(display, gc, white);
  }                             /* set background color */
  if (e && (e->type == Expose)){/* if called by Xt, */
    rc.x      = e->xexpose.x;   /* build and set the */
    rc.y      = e->xexpose.y;   /* clipping rectangle */
    rc.width  = e->xexpose.width;
    rc.height = e->xexpose.height;
    XSetClipRectangles(display, gc, 0, 0, &rc, 1, Unsorted);
  }

  /* --- draw window contents --- */
  draw_all();                   /* draw network and diagram(s) */
  XCopyArea(display, buffer, window, gc, 0, 0, INITXEXT, INITYEXT,0,0);
  sprintf(buf, "%d\n", mlp_epoch()); /* display the buffer and */
  mn_info(menu, buf);                /* show the current epoch */

  XFlush(display);              /* flush buffered drawing commands */
  XSetClipMask(display, gc, None);      /* and clear clipping mask */
}  /* redraw() */

/*--------------------------------------------------------------------*/

static void init (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Xor/Biimp/Function */
  int type;                     /* problem type */

  if      (!s || !c || (*c < 1))     type = MLP_XOR;
  else if (strcmp(*s, "func2") == 0) type = MLP_FUNC2;
  else if (strcmp(*s, "func1") == 0) type = MLP_FUNC1;
  else if (strcmp(*s, "biimp") == 0) type = MLP_BIIMP;
  else                               type = MLP_XOR;
  mlp_settype(type);            /* get and set the problem type, */
  mlp_init(range);              /* initialize the network weights, */
  diag = 0;                     /* toggle a full redraw, and */
  redraw(NULL,NULL,NULL,NULL);  /* redraw the window contents */
}  /* init() */

/*--------------------------------------------------------------------*/

static void exec (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Start/Stop */
  if (running) {                /* if running, remove procedure */
    mn_visible(menu, MN_ITEMS, 3);
    mn_enable (menu, MN_ITEMS, 3);
    running = 0; XtRemoveTimeOut(iid); }
  else {                        /* otherwise start training */
    mn_visible(menu, MN_ITEMS, 5);
    mn_enable (menu, MN_ITEMS, 4);
    running = 1; train(NULL, NULL);
  }
}  /* exec() */

/*--------------------------------------------------------------------*/

static void params (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Settings > Parameters... */
  char buf[32];                 /* buffer for numbers */

  if (!psh_params) {            /* if property sheet not created yet */
    psh_params = psh_create("paramDlg", w_top, 1);
    if (!psh_params) return;     /* create a property sheet */
    if ((psh_addline(psh_params, "lmode")               != 0)
    ||  (psh_additem(psh_params, "mode",   PSH_MENU, 0) != 0)
    ||  (psh_menuadd(psh_params, "mode",   "online")    != 0)
    ||  (psh_menuadd(psh_params, "mode",   "batch")     != 0)
    ||  (psh_addline(psh_params, "llrate")              != 0)
    ||  (psh_additem(psh_params, "lrate",  PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_params, "lmoment")             != 0)
    ||  (psh_additem(psh_params, "moment", PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_params, "lrange")              != 0)
    ||  (psh_additem(psh_params, "range",  PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_params, "ldelay")              != 0)
    ||  (psh_additem(psh_params, "delay",  PSH_EDIT, 0) != 0)) {
      psh_delete(psh_params); error(E_DIALOG); return; }
  }                             /* set browse button callback */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  psh_setval(psh_params, "mode",
             (char*)((mlp_getmode() == MLP_ONLINE) ? 0 : 1));
  sprintf(buf, "%g", mlp_getlrate());
  psh_setval(psh_params, "lrate",  buf);
  sprintf(buf, "%g", mlp_getmoment());
  psh_setval(psh_params, "moment", buf);
  sprintf(buf, "%g", range);
  psh_setval(psh_params, "range",  buf);
  sprintf(buf, "%g", delay);
  psh_setval(psh_params, "delay",  buf);
  psh_handle(psh_params, dcb_params); /* set values and handle dialog */
}  /* params() */

/*--------------------------------------------------------------------*/

static void weights (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Settings > Weights... */
  char buf[32];                 /* buffer for numbers */

  if (running) return;          /* if training is running, abort */
  if (!psh_wgts) {              /* if property sheet not created yet */
    psh_wgts = psh_create("weightDlg", w_top, 1);
    if (!psh_wgts) return;       /* create a property sheet */
    if ((psh_addline(psh_wgts, "llower")                != 0)
    ||  (psh_additem(psh_wgts, "lower_0",  PSH_EDIT, 0) != 0)
    ||  (psh_additem(psh_wgts, "lower_1",  PSH_EDIT, 0) != 0)
    ||  (psh_additem(psh_wgts, "lower_2",  PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_wgts, "lupper")                != 0)
    ||  (psh_additem(psh_wgts, "upper_0",  PSH_EDIT, 0) != 0)
    ||  (psh_additem(psh_wgts, "upper_1",  PSH_EDIT, 0) != 0)
    ||  (psh_additem(psh_wgts, "upper_2",  PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_wgts, "loutput")               != 0)
    ||  (psh_additem(psh_wgts, "output_0", PSH_EDIT, 0) != 0)
    ||  (psh_additem(psh_wgts, "output_1", PSH_EDIT, 0) != 0)
    ||  (psh_additem(psh_wgts, "output_2", PSH_EDIT, 0) != 0)) {
      psh_delete(psh_wgts); error(E_DIALOG); return; }
  }                             /* set browse button callback */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  sprintf(buf, "%g",  mlp_getwgt(MLP_HD1, MLP_IN1));
  psh_setval(psh_wgts, "lower_0",  buf);
  sprintf(buf, "%g",  mlp_getwgt(MLP_HD1, MLP_IN2));
  psh_setval(psh_wgts, "lower_1",  buf);
  sprintf(buf, "%g", -mlp_getwgt(MLP_HD1, MLP_BIAS));
  psh_setval(psh_wgts, "lower_2",  buf);
  sprintf(buf, "%g",  mlp_getwgt(MLP_HD2, MLP_IN1));
  psh_setval(psh_wgts, "upper_0",  buf);
  sprintf(buf, "%g",  mlp_getwgt(MLP_HD2, MLP_IN2));
  psh_setval(psh_wgts, "upper_1",  buf);
  sprintf(buf, "%g", -mlp_getwgt(MLP_HD2, MLP_BIAS));
  psh_setval(psh_wgts, "upper_2",  buf);
  sprintf(buf, "%g",  mlp_getwgt(MLP_OUT, MLP_IN1));
  psh_setval(psh_wgts, "output_0", buf);
  sprintf(buf, "%g",  mlp_getwgt(MLP_OUT, MLP_IN2));
  psh_setval(psh_wgts, "output_1", buf);
  sprintf(buf, "%g", -mlp_getwgt(MLP_OUT, MLP_BIAS));
  psh_setval(psh_wgts, "output_2", buf);
  psh_handle(psh_wgts, dcb_wgts);   /* set values and handle dialog */
}  /* weights() */

/*--------------------------------------------------------------------*/

static void about (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- 'About' dialog box */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  if (db_about(w_top, 0, dcb_nop) != 0)
    error(E_DIALOG);            /* show `about' dialog box */
}  /* about() */

/*----------------------------------------------------------------------
  Main Function
----------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main program */
  Arg       args[8];            /* widget arguments */
  int       n;                  /* argument counter */
  Dimension wd, ht;             /* width and height of window */
  Colormap  colmap;             /* color map of display */
  XColor    xc[9];              /* buffer for colors */

  XtSetLanguageProc(NULL, (XtLanguageProc)NULL, NULL);

  /* --- create main window --- */
  w_top = XtAppInitialize(&appctx, "xnn",
            NULL, 0, &argc, argv, fallback, NULL, 0);
  if (!w_top) error(E_INIT);    /* create toplevel widget */
  n = 0;                        /* build argument list */
  XtSetArg(args[n], XtNdefaultDistance, 0);    n++;
  XtSetArg(args[n], XtNresizable,       True); n++;
  w_main = XtCreateManagedWidget("main", formWidgetClass,
             w_top, args, n);   /* create main widget */
  if (!w_main) error(E_WIDGET);
  menu = mn_create(w_main);     /* create a menu bar and */
  if (!menu) error(E_NOMEM);    /* add titles and items */
  if ((mn_addtitle(menu, "file",     0, 0,          0)     != 0)
  ||  (mn_additem (menu, "load",     1, mcb_load,   0)     != 0)
  ||  (mn_additem (menu, "save",     0, mcb_save,   0)     != 0)
  ||  (mn_additem (menu, NULL,       0, 0,          0)     != 0)
  ||  (mn_additem (menu, "quit",     0, mcb_quit,   0)     != 0)
  ||  (mn_addtitle(menu, "actions",  0, 0,          0)     != 0)
  ||  (mn_additem (menu, "xor",      0, mcb_init, "xor")   != 0)
  ||  (mn_additem (menu, "biimp",    0, mcb_init, "biimp") != 0)
  ||  (mn_additem (menu, "func1",    0, mcb_init, "func1") != 0)
  ||  (mn_additem (menu, "func2",    0, mcb_init, "func2") != 0)
  ||  (mn_additem (menu, NULL,       0, 0,          0)     != 0)
  ||  (mn_additem (menu, "start",    2, mcb_start,  0)     != 0)
  ||  (mn_additem (menu, "stop",     4, mcb_start,  0)     != 0)
  ||  (mn_additem (menu, NULL,       0, 0,          0)     != 0)
  ||  (mn_additem (menu, "redraw",   0, mcb_redraw, 0)     != 0)
  ||  (mn_addtitle(menu, "settings", 0, 0,          0)     != 0)
  ||  (mn_additem (menu, "params",   0, mcb_params, 0)     != 0)
  ||  (mn_additem (menu, "weights",  1, mcb_wgts,   0)     != 0)
  ||  (mn_addtitle(menu, "help",     0, 0,          0)     != 0)
  ||  (mn_additem (menu, "about",    0, mcb_about,  0)     != 0)
  ||  (mn_addtitle(menu, NULL,       0, 0,          0)     != 0))
    error(E_WIDGET);
  mn_resize(menu, INITXEXT);    /* set width of menu bar */
  mn_visible(menu, MN_ITEMS, 3);/* show/hide      menu items */
  mn_enable (menu, MN_ITEMS, 3);/* enable/disable menu items */

  /* --- create viewport --- */
  n = 0;                        /* build argument list */
  XtSetArg(args[n], XtNfromVert,  mn_menubar(menu)); n++;
  XtSetArg(args[n], XtNleft,      XtChainLeft);      n++;
  XtSetArg(args[n], XtNtop,       XtChainTop);       n++;
  XtSetArg(args[n], XtNright,     XtChainRight);     n++;
  XtSetArg(args[n], XtNbottom,    XtChainBottom);    n++;
  XtSetArg(args[n], XtNresizable, True);             n++;
  w_view = XtCreateManagedWidget("view",
             coreWidgetClass, w_main, args, n);
  if (!w_view) error(E_WIDGET); /* create viewport */
  XtVaSetValues(w_view, XtNwidth,  INITXEXT,
                        XtNheight, INITYEXT, NULL);
  xoff = 0;                     /* set window size */
  yoff = INITYEXT -1;           /* and get offsets */
  XtOverrideTranslations(w_view,/* install redraw function */
    XtParseTranslationTable("<Expose>: redraw()"));

  /* --- install actions, accelerators and translations --- */
  XtAppAddActions(appctx, actions, XtNumber(actions));
  XtInstallAllAccelerators(w_main, w_main);
  
  /* --- show window --- */
  XtRealizeWidget(w_top);       /* show window on display */
  display = XtDisplay(w_top);   /* get display of top window */
  window  = XtWindow(w_view);   /* and drawable for window */
  XtVaGetValues(w_top, XtNwidth,   &wd, XtNheight  , &ht, NULL);
  XtVaSetValues(w_top, XtNminWidth, wd, XtNminHeight, ht,
                       XtNmaxWidth, wd, XtNmaxHeight, ht, NULL);
                                /* inhibit window resize */
  if (db_alert(w_top, dcb_nop, NULL) != 0)
    error(E_WIDGET);            /* create alert dialog box */

  /* --- allocate colors --- */
  screen = DefaultScreen(display);
  colmap = DefaultColormap(display, screen);
  white  = WhitePixel(display, screen);
  black  = BlackPixel(display, screen);
  if ((XAllocNamedColor(display, colmap, "grey",    xc+1, xc) == 0)
  ||  (XAllocNamedColor(display, colmap, "magenta", xc+2, xc) == 0)
  ||  (XAllocNamedColor(display, colmap, "cyan3",   xc+3, xc) == 0)
  ||  (XAllocNamedColor(display, colmap, "orange",  xc+4, xc) == 0)
  ||  (XAllocNamedColor(display, colmap, "yellow",  xc+5, xc) == 0)
  ||  (XAllocNamedColor(display, colmap, "green",   xc+6, xc) == 0)
  ||  (XAllocNamedColor(display, colmap, "blue",    xc+7, xc) == 0)
  ||  (XAllocNamedColor(display, colmap, "red",     xc+8, xc) == 0))
    error(E_COLOR);             /* allocate colors */
  grey   = xc[1].pixel; magenta = xc[2].pixel;
  cyan   = xc[3].pixel; orange  = xc[4].pixel;
  yellow = xc[5].pixel; green   = xc[6].pixel;    /* note color */
  blue   = xc[7].pixel; red     = xc[8].pixel;    /* pixel values */

  /* --- load font --- */
  font = XLoadQueryFont(display, "fixed");
  if (!font) error(E_FONT);     /* load a small fixed font */

  /* --- override window manager delete --- */
  XtOverrideTranslations(w_top, /* override window manager delete */
    XtParseTranslationTable("<Message>WM_PROTOCOLS: quit()"));
  wm_delwin = XInternAtom(XtDisplay(w_top), "WM_DELETE_WINDOW", False);
  XSetWMProtocols(XtDisplay(w_top), XtWindow(w_top), &wm_delwin, 1);

  srand((unsigned)time(NULL));  /* init. random number generator */
  init(NULL, NULL, NULL, NULL); /* initialize the neural network */

  /* --- start event loop --- */
  XtAppMainLoop(appctx);        /* enter main event loop */
  return 0;                     /* return `ok' */
}  /* main() */
